# Copyright (c) 2017 StackHPC Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os

import monasca_agent.collector.checks as checks

log = logging.getLogger(__name__)

# According to https://community.mellanox.com/docs/DOC-2572 these fields
# are divided by the number of lanes, so we need to multiply them by the lane
# count to get a number valid for the link as a whole.
_FIELDS_TO_MULTIPLY_BY_LANE_COUNT = {
    'port_rcv_data',
    'port_xmit_data'
}

_METRIC_NAME_PREFIX = "ibnet"
_IB_DEVICE_PATH = "/sys/class/infiniband/"
_IB_COUNTER_PATH = "ports/1/counters/"


class IBNetwork(checks.AgentCheck):
    def __init__(self, name, init_config, agent_config):
        super(IBNetwork, self).__init__(name, init_config, agent_config)

    @staticmethod
    def _get_lane_count():
        # It is possible that we could get the number of lanes from the driver,
        # for example:
        #
        # # cat /sys/class/infiniband/mlx5_0/ports/1/rate
        # 100 Gb/sec (4X EDR)
        #
        # However, according to the following PR this isn't expected to change:
        # https://github.com/prometheus/node_exporter/pull/579 so hard code it
        # for now.
        return 4

    def _normalise_counter(self, field, counter):
        if field in _FIELDS_TO_MULTIPLY_BY_LANE_COUNT:
            counter *= self._get_lane_count()
        return counter

    def _read_counter(self, device, field):
        counter_path = os.path.join(
            _IB_DEVICE_PATH, device, _IB_COUNTER_PATH, field)
        with open(counter_path) as f:
            counter = f.read()
        counter = int(counter.rstrip())
        counter = self._normalise_counter(field, counter)
        return counter

    @staticmethod
    def _get_devices():
        return os.listdir(_IB_DEVICE_PATH)

    @staticmethod
    def _get_fields(device):
        return os.listdir(os.path.join(
            _IB_DEVICE_PATH, device, _IB_COUNTER_PATH))

    def check(self, instance):
        dimensions = self._set_dimensions(None, instance)

        for device in self._get_devices():
            for field in self._get_fields(device):
                counter = self._read_counter(device, field)
                metric_name = '{0}.{1}'.format(_METRIC_NAME_PREFIX, field)
                self.rate(metric_name,
                          counter,
                          device_name=device,
                          dimensions=dimensions)
            log.debug('Collected network interface status for device {0}'.
                      format(device))



# (c) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 Fujitsu LIMITED
# Copyright 2017 SUSE Linux GmbH
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import pwd

from oslo_config import cfg
from oslo_utils import importutils

from monasca_setup import agent_config
from monasca_setup.detection import plugin
from monasca_setup.detection import utils


log = logging.getLogger(__name__)

# Directory to use for instance and metric caches (preferred tmpfs "/dev/shm")
cache_dir = "/dev/shm"
# Maximum age of instance cache before automatic refresh (in seconds)
nova_refresh = 60 * 60 * 4  # Four hours
# Probation period before metrics are gathered for a VM (in seconds)
vm_probation = 60 * 5  # Five minutes
# List of instance metadata keys to be sent as dimensions
# By default 'scale_group' metadata is used here for supporting auto
# scaling in Heat.
metadata = ['scale_group']
# Include scale group dimension for customer metrics.
customer_metadata = ['scale_group']
# List 'ping' commands (paths and parameters) in order of preference.
# The plugin will use the first functional command. 127.0.0.1 will be appended.
ping_options = [["/usr/bin/fping", "-n", "-c1", "-t250", "-q"],
                ["/sbin/fping", "-n", "-c1", "-t250", "-q"],
                ["/bin/ping", "-n", "-c1", "-w1", "-q"],
                ["/usr/bin/ping", "-n", "-c1", "-w1", "-q"]]
# Path to 'ip' command (needed to execute ping within network namespaces)
ip_cmd = "sudo /bin/ip"
# How many ping commands to run concurrently
default_max_ping_concurrency = 8
# Disk metrics can be collected at a larger interval than other vm metrics
default_disk_collection_period = 0
# VNIC metrics can be collected at a larger interval than other vm metrics
default_vnic_collection_period = 0

# Arguments which should be written as integers, not strings
INT_ARGS = ['disk_collection_period', 'vnic_collection_period',
            'max_ping_concurrency', 'nova_refresh', 'vm_probation']

_REQUIRED_OPTS = [
    {'opt': cfg.StrOpt('username'), 'group': 'keystone_authtoken'},
    {'opt': cfg.StrOpt('user_domain_name'), 'group': 'keystone_authtoken'},
    {'opt': cfg.StrOpt('password'), 'group': 'keystone_authtoken'},
    {'opt': cfg.StrOpt('project_name'), 'group': 'keystone_authtoken'},
    {'opt': cfg.StrOpt('project_domain_name'), 'group': 'keystone_authtoken'},
    {'opt': cfg.StrOpt('auth_url'), 'group': 'keystone_authtoken'}
]
"""Nova configuration opts required by this plugin"""


class Libvirt(plugin.Plugin):
    """Configures VM monitoring through Nova"""

    FAILED_DETECTION_MSG = 'libvirt plugin will not not be configured.'

    def _detect(self):
        """Set self.available True if the process and config file are detected
        """

        # NOTE(trebskit) bind each check we execute to another one
        # that way if X-one fails following won't be executed
        # and detection phase will end faster
        nova_proc = utils.find_process_name('nova-compute')
        has_deps = self.dependencies_installed() if nova_proc else None
        nova_conf = self._find_nova_conf(nova_proc) if has_deps else None
        has_cache_dir = self._has_cache_dir() if nova_conf else None
        agent_user = utils.get_agent_username() if has_cache_dir else None

        self.available = nova_conf and has_cache_dir
        if not self.available:
            if not nova_proc:
                detailed_message = '\tnova-compute process not found.'
                log.info('%s\n%s' % (detailed_message,
                                     self.FAILED_DETECTION_MSG))
            elif not has_deps:
                detailed_message = ('\tRequired dependencies were not found.\n'
                                    'Run pip install monasca-agent[libvirt] '
                                    'to install all dependencies.')
                log.warning('%s\n%s' % (detailed_message,
                                        self.FAILED_DETECTION_MSG))
            elif not has_cache_dir:
                detailed_message = '\tCache directory %s not found' % cache_dir
                log.warning('%s\n%s' % (detailed_message,
                                        self.FAILED_DETECTION_MSG))
            elif not nova_conf:
                detailed_message = ('\tnova-compute process was found, '
                                    'but it was impossible to '
                                    'read it\'s configuration.')
                log.warning('%s\n%s' % (detailed_message,
                                        self.FAILED_DETECTION_MSG))
        else:
            self.nova_conf = nova_conf
            self._agent_user = agent_user

    def build_config(self):
        """Build the config as a Plugins object and return back.
        """
        config = agent_config.Plugins()
        init_config = self._get_init_config()

        self._configure_ping(init_config)

        # Handle monasca-setup detection arguments, which take precedence
        if self.args:
            for arg in self.args:
                if arg in INT_ARGS:
                    value = self.args[arg]
                    try:
                        init_config[arg] = int(value)
                    except ValueError:
                        log.warn("\tInvalid integer value '{0}' for parameter {1}, ignoring value"
                                 .format(value, arg))
                else:
                    init_config[arg] = self.literal_eval(self.args[arg])

        config['libvirt'] = {'init_config': init_config,
                             'instances': []}

        return config

    def _configure_ping(self, init_config):
        if self._agent_user is None:
            log.warn("\tUnable to determine agent user. Skipping ping checks.")
            return

        client = importutils.try_import('neutronclient.v2_0.client',
                                        False)
        if not client:
            log.warning(
                '\tpython-neutronclient module missing, '
                'required for ping checks.')
            return

        # Look for the best ping command
        for ping_cmd in ping_options:
            if os.path.isfile(ping_cmd[0]):
                init_config[
                    'ping_check'] = "{0} netns exec NAMESPACE {1}".format(
                        ip_cmd,
                        ' '.join(ping_cmd))
                log.info(
                    "\tEnabling ping checks using {0}".format(ping_cmd[0]))
                break
        if init_config['ping_check'] is False:
            log.warn('\tUnable to find suitable ping command, '
                     'disabling ping checks.')

    def dependencies_installed(self):
        return importutils.try_import('novaclient.client', False)

    def _get_init_config(self):
        keystone_auth_section = self.nova_conf['keystone_authtoken']
        init_config = {
            'cache_dir': cache_dir,
            'nova_refresh': nova_refresh,
            'metadata': metadata,
            'vm_probation': vm_probation,
            'customer_metadata': customer_metadata,
            'max_ping_concurrency': default_max_ping_concurrency,
            'disk_collection_period': default_disk_collection_period,
            'vnic_collection_period': default_vnic_collection_period,
            'vm_cpu_check_enable': True,
            'vm_disks_check_enable': True,
            'vm_network_check_enable': True,
            'vm_ping_check_enable': True,
            'vm_extended_disks_check_enable': False,
            'ping_check': False,
            'username': keystone_auth_section['username'],
            'user_domain_name': keystone_auth_section['user_domain_name'],
            'password': keystone_auth_section['password'],
            'project_name': keystone_auth_section['project_name'],
            'project_domain_name': keystone_auth_section['project_domain_name'],
            'auth_url': keystone_auth_section['auth_url']
        }
        return init_config

    @staticmethod
    def _has_cache_dir():
        return os.path.isdir(cache_dir)

    @staticmethod
    def _find_nova_conf(nova_process):
        try:
            nova_cmd = nova_process.as_dict(['cmdline'])['cmdline']
            return utils.load_oslo_configuration(from_cmd=nova_cmd,
                                                 in_project='nova',
                                                 for_opts=_REQUIRED_OPTS)
        except cfg.Error:
            log.exception('Failed to load nova configuration')
        return None

    @staticmethod
    def _get_user_uid_gid(username):
        stat = pwd.getpwnam(username)
        uid = stat.pw_uid
        gid = stat.pw_gid
        return uid, gid



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import configparser
import logging
import os
import re

from oslo_config import cfg
from oslo_utils import importutils

from monasca_setup import agent_config
from monasca_setup import detection
from monasca_setup.detection import utils

log = logging.getLogger(__name__)

# Maximum time to wait before updating neutron router cache(in seconds)
neutron_refresh = 60 * 60 * 4  # Four hours
# Directory to use for metric caches
cache_dir = "/dev/shm"
# ovs-vsctl command needs sudo privileges to connect to
# /var/run/openvswitch/db.sock which is created by ovsdb-server.
ovs_cmd = "sudo /usr/bin/ovs-vsctl"
# If set, will submit network metrics in bits
network_use_bits = False
# Regular expression for interface types
included_interface_re = 'qg.*|vhu.*|sg.*'
# If set, will submit raw counters from ovs-vsctl command output for the given
# network interface
use_absolute_metrics = True
# If set, will submit the rate metrics
use_rate_metrics = True
# If set, will submit the health metrics
use_health_metrics = True
# If set, router max capacity metrics will be published
publish_router_capacity = False
# Acceptable arguments
acceptable_args = ['username', 'password', 'project_name',
                   'user_domain_name', 'user_domain_id',
                   'project_domain_name', 'project_domain_id'
                   'auth_url', 'cache_dir', 'neutron_refresh', 'ovs_cmd',
                   'network_use_bits', 'check_router_ha', 'region_name',
                   'included_interface_re', 'conf_file_path', 'use_absolute_metrics',
                   'use_rate_metrics', 'use_health_metrics', 'publish_router_capacity']
# Arguments which must be ignored if provided
ignorable_args = ['username', 'password', 'project_name', 'user_domain_name',
                  'user_domain_id', 'project_domain_name', 'project_domain_id',
                  'auth_url', 'region_name', 'conf_file_path']


class Ovs(detection.Plugin):
    """Detect OVS daemons and setup configuration to monitor

    """

    PROC_NAME = 'neutron-openvsw'
    """Name of the ovs process to look for"""
    REQUIRED_CONF_SECTIONS = 'keystone_authtoken',
    """Tuple of sections that must be part of neutron configuration file"""

    def _detect(self):
        process_exist = utils.find_process_cmdline(Ovs.PROC_NAME)
        has_dependencies = self.dependencies_installed()
        neutron_conf = self._get_ovs_config_file() if process_exist else ''
        neutron_conf_exists = os.path.isfile(neutron_conf)
        neutron_conf_valid = (neutron_conf_exists and
                              self._is_neutron_conf_valid(neutron_conf))

        self.available = (process_exist is not None and
                          neutron_conf_valid and has_dependencies)
        self.cmd = ''
        if process_exist:
            self.cmd = process_exist.as_dict(attrs=['cmdline'])['cmdline']

        if not self.available:
            if not process_exist:
                log.info('OVS daemon process [%s] does not exist.',
                         Ovs.PROC_NAME)
            elif not neutron_conf_exists:
                log.warning(('OVS daemon process exists but configuration '
                             'file was not found. Path to file does not exist '
                             'as a process parameter or was not '
                             'passed via args.'))
            elif not neutron_conf_valid:
                log.error(('OVS daemon process exists, configuration file was '
                           'found but it looks like it does not contain '
                           'one of following sections=%s. '
                           'Check if neutron was not configured to load '
                           'configuration from /etc/neutron/neutron.conf.d/.'),
                          Ovs.REQUIRED_CONF_SECTIONS)
                # NOTE(trebskit) the problem with that approach is that
                # each setting that Ovs plugin require might be scattered
                # through multiple files located inside
                # /etc/neutron/neutron.conf.d/
                # not to mention that it is still possible to have
                # yet another configuration file passed as neutron CLI
                # argument
            elif not has_dependencies:
                log.error(('OVS daemon process exists but required '
                           'dependencies were not found.\n'
                           'Run pip install monasca-agent[ovs] '
                           'to install all dependencies.'))
        else:
            for_opts = [{'opt': cfg.StrOpt('region', default='RegionOne'),
                         'group': 'service_auth'},
                        {'opt': cfg.StrOpt('region_name'), 'group': 'nova'},
                        {'opt': cfg.StrOpt('nova_region_name'),
                         'group': 'DEFAULT'},
                        {'opt': cfg.StrOpt('username'),
                         'group': 'keystone_authtoken'},
                        {'opt': cfg.StrOpt('password'),
                         'group': 'keystone_authtoken'},
                        {'opt': cfg.StrOpt('project_name'),
                         'group': 'keystone_authtoken'},
                        {'opt': cfg.StrOpt('user_domain_name'),
                         'group': 'keystone_authtoken'},
                        {'opt': cfg.StrOpt('user_domain_id'),
                         'group': 'keystone_authtoken'},
                        {'opt': cfg.StrOpt('project_domain_name'),
                         'group': 'keystone_authtoken'},
                        {'opt': cfg.StrOpt('project_domain_id'),
                         'group': 'keystone_authtoken'},
                        {'opt': cfg.StrOpt('auth_url'),
                         'group': 'keystone_authtoken'},
                        {'opt': cfg.StrOpt('identity_uri'),
                         'group': 'keystone_authtoken'}]
            self.conf = utils.load_oslo_configuration(from_cmd=self.cmd,
                                                      in_project='neutron',
                                                      for_opts=for_opts
                                                      )
            log.info("\tUsing neutron configuration file {0} for detection".format(neutron_conf))
            self.neutron_conf = neutron_conf

    def _is_neutron_conf_valid(self, conf_file):
        neutron_cfg = configparser.SafeConfigParser()
        neutron_cfg.read(conf_file)

        for section in self.REQUIRED_CONF_SECTIONS:
            if not neutron_cfg.has_section(section=section):
                return False

        return True

    def _get_ovs_config_file(self):
        neutron_conf = None
        if self.args:
            for arg in self.args:
                if arg == 'conf_file_path':
                    neutron_conf = self.args[arg]
        # Try to detect the location of the Neutron configuration file.
        # Walk through the list of processes, searching for 'neutron'
        # process with 'neutron.conf' inside one of the parameters.
        if not neutron_conf:
            proc = utils.find_process_name(Ovs.PROC_NAME)
            proc_dict = proc.as_dict(attrs=['cmdline'])
            cmd = proc_dict['cmdline']
            neutron_config_params = [param for param in cmd if
                                     'neutron.conf' in param]
            if neutron_config_params:
                if '=' in neutron_config_params[0]:
                    neutron_conf = neutron_config_params[0].split('=')[1]
                else:
                    neutron_conf = neutron_config_params[0]
        return neutron_conf

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        config = agent_config.Plugins()
        log.info("\tUsing neutron configuration file {0} and configuration dir"
                 " {1}".format(self.conf.default_config_files, self.conf.default_config_dirs))
        cfg_section = 'keystone_authtoken'
        cfg_needed = {'username': 'username',
                      'password': 'password',
                      'project_name': 'project_name'}
        cfg_domain = ['user_domain_name', 'user_domain_id',
                      'project_domain_name', 'project_domain_id']

        # Start with plugin-specific configuration parameters
        init_config = {'cache_dir': cache_dir,
                       'neutron_refresh': neutron_refresh,
                       'ovs_cmd': ovs_cmd,
                       'network_use_bits': network_use_bits,
                       'included_interface_re': included_interface_re,
                       'use_absolute_metrics': use_absolute_metrics,
                       'use_rate_metrics': use_rate_metrics,
                       'use_health_metrics': use_health_metrics,
                       'publish_router_capacity': publish_router_capacity}

        for option in cfg_needed:
            init_config[option] = self.get_option(cfg_section, cfg_needed[option])

        for option in cfg_domain:
            if self.has_option(cfg_section, option):
                init_config[option] = self.get_option(cfg_section, option)

        # Create an identity URI (again, slightly different for Devstack)
        if self.has_option(cfg_section, 'auth_url'):
            init_config['auth_url'] = self.get_option(cfg_section, 'auth_url')
        else:
            init_config['auth_url'] = self.get_option(cfg_section, 'identity_uri')

        # Create an region_name (again, slightly different for Devstack)
        if self.has_option('service_auth', 'region'):
            init_config['region_name'] = str(self.get_option('service_auth', 'region'))
        else:
            try:
                init_config['region_name'] = str(self.get_option('nova', 'region_name'))
            except configparser.NoOptionError:
                log.debug(('Option region_name was not found in nova '
                           'section, nova_region_name option from '
                           'DEFAULT section will be used.'))
                init_config['region_name'] = str(self.get_option('DEFAULT',
                                                                 'nova_region_name'))

        # Handle monasca-setup detection arguments, which take precedence
        if self.args:
            for arg in self.args:
                if arg in acceptable_args and arg not in ignorable_args:
                    if arg == 'included_interface_re':
                        try:
                            re.compile(self.args[arg])
                        except re.error as e:
                            exception_msg = (
                                "Invalid regular expression given for "
                                "'included_interface_re'. {0}.".format(e))
                            log.exception(exception_msg)
                            raise Exception(exception_msg)

                    init_config[arg] = self.literal_eval(self.args[arg])
                elif arg in ignorable_args:
                    log.warn("Argument '{0}' is ignored; Fetching {0} from "
                             "neutron configuration file.".format(arg))
                else:
                    log.warn("Invalid argument '{0}' "
                             "has been provided!!!".format(arg))

        config['ovs'] = {'init_config': init_config,
                         'instances': []}

        return config

    def dependencies_installed(self):
        return (importutils.try_import('neutronclient', False) and
                importutils.try_import('novaclient', False))

    def has_option(self, section, option):
        return option in self.conf.get(section)

    def get_option(self, section, option):
        return self.conf.get(section).get(option)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import monasca_setup.detection
import socket
import unittest


from monasca_agent.common.keystone import Keystone

port_used = 0


class TestKeystone(unittest.TestCase):

    def setUp(self):
        global port_used
        # Create a server socket so the htto check config gets created
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            self.s.bind(('', 0))
            self.s.listen(1)
        except socket.error as msg:
            raise Exception('Bind failed. Message ' + msg[1])
        (host, port_used) = self.s.getsockname()
        
    def tearDown(self):
        self.s.close()

    def test_no_override_(self):
        """ Test setting values with no overrides works as expected
        """
        args = None
        test_plugin = TestPlugin('.', args=args)
        test_plugin._detect()
        config = test_plugin.build_config()
        http_instance = config['http_check']['instances'][0]
        url = 'http://localhost:{0}/healthcheck'.format(port_used)
        self.assertEqual(http_instance['url'], url)
        self.assertEqual(http_instance['match_pattern'], '.*OK.*')

        processes = config['process']['instances']
        self.assertEqual(processes[0]['search_string'], ['testr'])

    def test_override_values(self):
        """ Test overriding values using args works
        """
        url = 'http://localhost:{0}/othercheck'.format(port_used)
        pattern = 'CHECK.*'
        args = 'process_names=tox,testr service_api_url='
        args += ' service_api_url=' + url
        args += ' search_pattern=' + pattern
        test_plugin = TestPlugin('.', args=args)
        test_plugin._detect()
        config = test_plugin.build_config()

        http_instance = config['http_check']['instances'][0]
        self.assertEqual(http_instance['url'], url)
        self.assertEqual(http_instance['match_pattern'], pattern)

        processes = config['process']['instances']
        self.assertEqual(processes[0]['search_string'], ['tox'])
        self.assertEqual(processes[1]['search_string'], ['testr'])


class TestPlugin(monasca_setup.detection.ServicePlugin):

    """Test Plugin

    """

    def __init__(self, template_dir, overwrite=True, args=None):
        url = 'http://localhost:{0}/healthcheck'.format(port_used)
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': 'os-testr',
            'process_names': ['testr'],
            'service_api_url': url,
            'search_pattern': '.*OK.*'
        }

        super(TestPlugin, self).__init__(service_params)



# (C) Copyright 2016 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import unittest

import monasca_agent.common.util as util
from monasca_agent.collector.checks_d.host_alive import HostAlive

HOST_ALIVE_STATUS = 'host_alive_status'
SUCCESS = 0
FAILURE = 1


class TestHostAlive(unittest.TestCase):

    def setUp(self):
        unittest.TestCase.setUp(self)
        init_config = {}
        agent_config = {}
        self._host_alive = HostAlive('TestHostAlive', init_config, agent_config)
        self._gauge = mock.Mock()
        self._host_alive.gauge = self._gauge
        self._host_name = 'monasca'
        self._instance = {'host_name': self._host_name,
                    'alive_test': 'ping'}
        self._base_dimensions = {
            'test_type': 'ping',
            'hostname': self._host_name,
            'observer_host': util.get_hostname()
        }

    def _run_check(self, host_name, instance, ping_result):
        mock_ping = mock.Mock(return_value=ping_result)
        self._host_alive._test_ping = mock_ping
        self._host_alive._check(instance)
        mock_ping.assert_called_with(host_name, None)

    def test_host_is_alive(self):
        ping_result = (True, None)
        self._run_check(self._host_name, self._instance, ping_result)
        self._gauge.assert_called_with(HOST_ALIVE_STATUS,
                       SUCCESS,
                       dimensions=self._base_dimensions)

    def test_host_is_dead(self):
        error_message = '''I'm not dead yet'''
        self._run_check(self._host_name, self._instance,
                       (False, error_message))

        self._gauge.assert_called_with('host_alive_status',
                       FAILURE,
                       dimensions=self._base_dimensions,
                       value_meta={'error': error_message})

    def test_host_is_alive_with_target_hostname(self):
        check_name = 'otherMonasca'
        self._instance['target_hostname'] = check_name
        self._run_check(check_name, self._instance, (True, None))
        self._base_dimensions['target_hostname'] = check_name
        self._gauge.assert_called_with(HOST_ALIVE_STATUS,
                       SUCCESS,
                       dimensions=self._base_dimensions)

    def test_host_is_alive_with_dup_target_hostname(self):
        host_name = 'monasca'
        self._instance['target_hostname'] = host_name
        self._run_check(host_name, self._instance, (True, None))
        self._gauge.assert_called_with(HOST_ALIVE_STATUS,
                       SUCCESS,
                       dimensions=self._base_dimensions)



# (C) Copyright 2015-2017 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest


import monasca_agent.common.aggregator as aggregator
import monasca_agent.common.metrics as metrics_pkg

import monasca_common.validation.metrics as metric_validator

# a few valid characters to test
valid_name_chars = ".'_-"
invalid_name_chars = metric_validator.INVALID_CHARS

# a few valid characters to test
valid_dimension_chars = " .'_-"
invalid_dimension_chars = metric_validator.INVALID_CHARS


class TestMetricsAggregator(unittest.TestCase):
    def setUp(self):
        self.aggregator = aggregator.MetricsAggregator("Foo")

    def submit_metric(self, name, value,
                      dimensions=None,
                      value_meta=None,
                      exception=None):
        if exception:
            with self.assertRaises(exception):
                self.aggregator.submit_metric(name,
                                              value,
                                              metrics_pkg.Gauge,
                                              dimensions=dimensions,
                                              delegated_tenant=None,
                                              hostname=None,
                                              device_name=None,
                                              value_meta=value_meta)
        else:
            self.aggregator.submit_metric(name,
                                          value,
                                          metrics_pkg.Gauge,
                                          dimensions=dimensions,
                                          delegated_tenant=None,
                                          hostname=None,
                                          device_name=None,
                                          value_meta=value_meta)

    def testValidMetric(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta)

    def testValidMetricUnicodeDimensionValue(self):
        dimensions = {chr(2440): 'B', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta)

    def testValidMetricUnicodeDimensionKey(self):
        dimensions = {'A': 'B', 'B': chr(920), 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta)

    def testValidMetricUnicodeMetricName(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric(chr(6021),
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta)

    def testInvalidMetricName(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("TooLarge" * 255,
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidMetricName)

    def testInvalidMetricNameEmpty(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric('',
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidMetricName)

    def testInvalidMetricNameNonStr(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric(133,
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidMetricName)

    def testInvalidMetricRestrictedCharacters(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric('"Foo"',
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidMetricName)

    def testInvalidDimensionEmptyKey(self):
        dimensions = {'A': 'B', '': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidDimensionKey)

    def testInvalidDimensionEmptyValue(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': ''}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidDimensionValue)

    def testInvalidDimensionNonStrKey(self):
        dimensions = {'A': 'B', 4: 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidDimensionKey)

    def testInvalidDimensionNonStrValue(self):
        dimensions = {'A': 13.3, 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidDimensionValue)

    def testInvalidDimensionKeyLength(self):
        dimensions = {'A'*256: 'B', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidDimensionKey)

    def testInvalidDimensionValueLength(self):
        dimensions = {'A': 'B', 'B': 'C'*256, 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidDimensionValue)

    def testValidDimensionKeyParenthesesCharacter(self):
        dimensions = {'A': 'B', 'B': 'C', '(D)': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta)

    def testInvalidDimensionValueRestrictedCharacters(self):
        dimensions = {'A': 'B;', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidDimensionValue)

    def testInvalidDimensionKeyLeadingUnderscore(self):
        dimensions = {'_A': 'B', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           5,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidDimensionKey)

    def testInvalidValue(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        value_meta = {"This is a test": "test, test, test"}
        self.submit_metric("Foo",
                           "value",
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidValue)

    def testValidNameChars(self):
        for c in valid_name_chars:
            self.submit_metric('test{}counter'.format(c), 2,
                               dimensions={"test-key": "test-value"})

    def testInvalidNameChars(self):
        for c in invalid_name_chars:
            self.submit_metric('test{}counter'.format(c), 2,
                               dimensions={"test-key": "test-value"},
                               exception=metric_validator.InvalidMetricName)

    def testValidDimensionChars(self):
        for c in valid_dimension_chars:
            self.submit_metric('test-counter', 2,
                               dimensions={"test{}key".format(c): "test{}value".format(c)})

    def testInvalidDimensionChars(self):
        for c in invalid_dimension_chars:
            self.submit_metric('test-counter', 2,
                               dimensions={'test{}key'.format(c): 'test-value'},
                               exception=metric_validator.InvalidDimensionKey)
            self.submit_metric('test-counter', 2,
                               dimensions={'test-key': 'test{}value'.format(c)},
                               exception=metric_validator.InvalidDimensionValue)

    def testTooManyValueMeta(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        value_meta = {}
        for i in range(0, 17):
            value_meta['key{}'.format(i)] = 'value{}'.format(i)
        self.submit_metric("Foo",
                           2,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidValueMeta)

    def testEmptyValueMetaKey(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        value_meta = {'': 'BBB'}
        self.submit_metric("Foo",
                           2,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidValueMeta)

    def testTooLongValueMetaKey(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        key = "K"
        for i in range(0, metric_validator.VALUE_META_NAME_MAX_LENGTH):
            key = "{}{}".format(key, "1")
        value_meta = {key: 'BBB'}
        print(key)
        self.submit_metric("Foo",
                           2,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidValueMeta)

    def testTooLargeValueMeta(self):
        dimensions = {'A': 'B', 'B': 'C', 'D': 'E'}
        value_meta_value = ""
        num_value_meta = 10
        for i in range(0, metric_validator.VALUE_META_VALUE_MAX_LENGTH//num_value_meta):
            value_meta_value = '{}{}'.format(value_meta_value, '1')

        value_meta = {}
        for i in range(0, num_value_meta):
            value_meta['key{}'.format(i)] = value_meta_value
        self.submit_metric("Foo",
                           2,
                           dimensions=dimensions,
                           value_meta=value_meta,
                           exception=metric_validator.InvalidValueMeta)



# (C) Copyright 2015,2017-2018 Hewlett Packard Enterprise Development LP
# (C) Copyright 2017 KylinCloud
# Copyright 2018 OP5 AB
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import base64
import json
import logging
import math
from numbers import Number
import os
import re
import requests

from monasca_agent.common import exceptions
from monasca_agent.common import keystone

log = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 20


def add_basic_auth(request, username, password):
    """A helper to add basic authentication to a urllib2 request.

    We do this across a variety of checks so it's good to have this in one place.
    """
    auth_str = base64.encodestring('%s:%s' % (username, password)).strip()
    request.add_header('Authorization', 'Basic %s' % auth_str)
    return request


def get_tenant_name(tenants, tenant_id):
    tenant_name = None
    for tenant in tenants:
        if tenant.id == tenant_id:
            tenant_name = tenant.name
            break
    return tenant_name


def get_tenant_list(config, log):
    tenants = []
    try:
        log.debug("Retrieving Keystone tenant list")
        client = keystone.get_client(**config)
        if 'v2' in client.__module__:
            tenants = client.tenants.list()
        else:
            tenants = client.projects.list()
    except Exception as e:
        msg = "Unable to get tenant list from keystone: {0}"
        log.error(msg.format(e))

    return tenants


def convert_memory_string_to_bytes(memory_string):
    """Conversion from memory represented in string format to bytes"""
    if "m" in memory_string:
        memory = float(memory_string.split('m')[0])
        return memory / 1000
    elif "K" in memory_string:
        memory = float(memory_string.split('K')[0])
        return _compute_memory_bytes(memory_string, memory, 1)
    elif "M" in memory_string:
        memory = float(memory_string.split('M')[0])
        return _compute_memory_bytes(memory_string, memory, 2)
    elif "G" in memory_string:
        memory = float(memory_string.split('G')[0])
        return _compute_memory_bytes(memory_string, memory, 3)
    elif "T" in memory_string:
        memory = float(memory_string.split('T')[0])
        return _compute_memory_bytes(memory_string, memory, 4)
    else:
        return float(memory_string)


def _compute_memory_bytes(memory_string, memory, power):
    if "i" in memory_string:
        return memory * math.pow(1024, power)
    return memory * math.pow(1000, power)


class KubernetesConnector(object):
    """Class for connecting to Kubernetes API from within a container running
    in a Kubernetes environment
    """
    CACERT_PATH = '/var/run/secrets/kubernetes.io/serviceaccount/ca.crt'
    TOKEN_PATH = '/var/run/secrets/kubernetes.io/serviceaccount/token'  # nosec B105

    def __init__(self, connection_timeout):
        self.api_url = None
        self.api_verify = None
        self.api_request_header = None
        if connection_timeout is None:
            self.connection_timeout = DEFAULT_TIMEOUT
        else:
            self.connection_timeout = connection_timeout
        self._set_kubernetes_api_connection_info()
        self._set_kubernetes_service_account_info()

    def _set_kubernetes_api_connection_info(self):
        """Set kubernetes API string from default container environment
        variables
        """
        api_host = os.environ.get('KUBERNETES_SERVICE_HOST', "kubernetes.default")
        api_port = os.environ.get('KUBERNETES_SERVICE_PORT', "443")
        self.api_url = "https://{}:{}".format(api_host, api_port)

    def _set_kubernetes_service_account_info(self):
        """Set cert and token info to included on requests to the API"""
        try:
            with open(self.TOKEN_PATH) as token_file:
                token = token_file.read()
        except Exception as e:
            log.error("Unable to read token - {}. Defaulting to using no token".format(e))
            token = None
        self.api_request_header = {'Authorization': 'Bearer {}'.format(token)} if token else None
        self.api_verify = self.CACERT_PATH if os.path.exists(self.CACERT_PATH) else False

    def get_agent_pod_host(self, return_host_name=False):
        """Obtain host the agent is running on in Kubernetes.
        Used when trying to connect to services running on the node (Kubelet, cAdvisor)
        """
        # Get pod name and namespace from environment variables
        pod_name = os.environ.get("AGENT_POD_NAME")
        pod_namespace = os.environ.get("AGENT_POD_NAMESPACE")
        if not pod_name:
            raise exceptions.MissingEnvironmentVariables(
                "pod_name is not set as environment variables cannot derive"
                " host from Kubernetes API")
        if not pod_namespace:
            raise exceptions.MissingEnvironmentVariables(
                "pod_namespace is not set as environment variables cannot "
                "derive host from Kubernetes API")
        pod_url = "/api/v1/namespaces/{}/pods/{}".format(pod_namespace, pod_name)
        try:
            agent_pod = self.get_request(pod_url)
        except Exception as e:
            exception_message = "Could not get agent pod from Kubernetes API" \
                                " to get host IP with error - {}".format(e)
            log.exception(exception_message)
            raise exceptions.KubernetesAPIConnectionError(exception_message)
        if not return_host_name:
            return agent_pod['status']['hostIP']
        else:
            return agent_pod['spec']['nodeName']

    def get_request(self, request_endpoint, as_json=True, retried=False):
        """Sends request to Kubernetes API with given endpoint.
        Will retry the request once, with updated token/cert, if unauthorized.
        """
        api_url = self.api_url
        if api_url[-1] == '/':
            api_url = api_url[:-1]

        if request_endpoint[0] == '/':
            request_endpoint = request_endpoint[1:]

        request_url = "{}/{}".format(api_url, request_endpoint)
        result = requests.get(request_url,
                              timeout=self.connection_timeout,
                              headers=self.api_request_header,
                              verify=self.api_verify)
        if result.status_code >= 300:
            if result.status_code == 401 and not retried:
                log.info("Could not authenticate with Kubernetes API at the"
                         " first time. Rereading in cert and token.")
                self._set_kubernetes_service_account_info()
                return self.get_request(request_endpoint, as_json=as_json,
                                        retried=True)
            exception_message = "Could not obtain data from {} with the " \
                                "given status code {} and return text {}".\
                format(request_url, result.status_code, result.text)
            raise exceptions.KubernetesAPIConnectionError(exception_message)
        return result.json() if as_json else result


class DynamicCheckHelper(object):
    """Supplements existing check class with reusable functionality to transform
    third-party metrics into Monasca ones in a configurable way
    """

    COUNTERS_KEY = 'counters'
    RATES_KEY = 'rates'
    GAUGES_KEY = 'gauges'
    CONFIG_SECTIONS = [GAUGES_KEY, RATES_KEY, COUNTERS_KEY]

    GAUGE = 0
    RATE = 1
    COUNTER = 2
    SKIP = 3
    METRIC_TYPES = [GAUGE, RATE, COUNTER]

    DEFAULT_GROUP = ""

    class MetricSpec(object):
        """Describes how to filter and map input metrics to Monasca metrics
        """

        def __init__(self, metric_type, metric_name):
            """Construct a metric-specification
            :param metric_type: one of GAUGE, RATE, COUNTER, SKIP
            :param metric_name: normalized name of the metric as reported to Monasca
            """
            self.metric_type = metric_type
            self.metric_name = metric_name

    @staticmethod
    def _normalize_dim_value(value):
        """Normalize an input value

        * Replace \\x?? values with _
        * Replace illegal characters
          - according to ANTLR grammar:
          ( '}' | '{' | '&' | '|' | '>' | '<' | '=' | ',' | ')' | '(' | ' ' | '"' )
          - according to Python API validation: "<>={}(),\"\\\\|;&"
        * Truncate to 255 chars
        :param value: input value
        :return: valid dimension value
        """

        return re.sub(r'[|\\;,&=\']', '-',
                      re.sub(r'[(}>]', ']',
                             re.sub(r'[({<]', '[', value.replace(r'\x2d', '-').
                                    replace(r'\x7e', '~'))))[:255]

    class DimMapping(object):
        """Describes how to transform dictionary like metadata attached to a metric into Monasca dimensions
        """

        def __init__(self, dimension, regex='(.*)', separator=None):
            """C'tor
            :param dimension to be mapped to
            :param regex: regular expression used to extract value from source value
            :param separator: used to concatenate match-groups
            """
            self.dimension = dimension
            self.regex = regex
            self.separator = separator
            self.cregex = re.compile(regex) if regex != '(.*)' else None

        def map_value(self, source_value):
            """Transform source value into target dimension value
            :param source_value: label value to transform
            :return: transformed dimension value or None if the regular
            expression did not match. An empty result (caused by the regex
            having no match-groups) indicates that the label is used for filtering
            but not mapped to a dimension.
            """
            if self.cregex:
                match_groups = self.cregex.match(source_value)
                if match_groups:
                    return DynamicCheckHelper._normalize_dim_value(
                        self.separator.join(match_groups.groups()))
                else:
                    return None
            else:
                return DynamicCheckHelper._normalize_dim_value(source_value)

    @staticmethod
    def _build_dimension_map(config):
        """Builds dimension mappings for the given configuration element
        :param config: 'mappings' element of config
        :return: dictionary mapping source labels to applicable DimMapping objects
        """
        result = {}
        for dim, spec in config.get('dimensions', {}).items():
            if isinstance(spec, dict):
                label = spec.get('source_key', dim)
                sepa = spec.get('separator', '-')
                regex = spec.get('regex', '(.*)')
            else:
                label = spec
                regex = '(.*)'
                sepa = None

            # note: source keys can be mapped to multiple dimensions
            arr = result.get(label, [])
            mapping = DynamicCheckHelper.DimMapping(dimension=dim, regex=regex, separator=sepa)
            arr.append(mapping)
            result[label] = arr

        return result

    def __init__(self, check, prefix=None, default_mapping=None):
        """C'tor
        :param check: Target check instance
        """
        self._check = check
        self._prefix = prefix
        self._groups = {}
        self._metric_map = {}
        self._dimension_map = {}
        self._metric_cache = {}
        self._grp_metric_map = {}
        self._grp_dimension_map = {}
        self._grp_metric_cache = {}
        self._metric_to_group = {}
        for inst in self._check.instances:
            iname = inst['name']
            mappings = inst.get('mapping', default_mapping)
            if mappings:
                # build global name filter and rate/gauge assignment
                self._metric_map[iname] = mappings
                self._metric_cache[iname] = {}
                # build global dimension map
                self._dimension_map[iname] = DynamicCheckHelper._build_dimension_map(mappings)
                # check if groups are used
                groups = mappings.get('groups')
                self._metric_to_group[iname] = {}
                self._groups[iname] = []
                if groups:
                    self._groups[iname] = list(groups.keys())
                    self._grp_metric_map[iname] = {}
                    self._grp_metric_cache[iname] = {}
                    self._grp_dimension_map[iname] = {}
                    for grp, gspec in groups.items():
                        self._grp_metric_map[iname][grp] = gspec
                        self._grp_metric_cache[iname][grp] = {}
                        self._grp_dimension_map[iname][grp] =\
                            DynamicCheckHelper._build_dimension_map(gspec)
                    # add the global mappings as pseudo group, so that it is considered when
                    # searching for metrics
                    self._groups[iname].append(DynamicCheckHelper.DEFAULT_GROUP)
                    self._grp_metric_map[iname][DynamicCheckHelper.DEFAULT_GROUP] =\
                        self._metric_map[iname]
                    self._grp_metric_cache[iname][DynamicCheckHelper.DEFAULT_GROUP] =\
                        self._metric_cache[iname]
                    self._grp_dimension_map[iname][DynamicCheckHelper.DEFAULT_GROUP] =\
                        self._dimension_map[iname]

            else:
                raise exceptions.CheckException(
                    'instance %s is not supported: no element "mapping" found!', iname)

    def _get_group(self, instance, metric):
        """Search the group for a metric. Can be used only when metric names unambiguous across groups.

        :param metric: input metric
        :return: group name or None (if no group matches)
        """
        iname = instance['name']
        group = self._metric_to_group[iname].get(metric)
        if group is None:
            for g in self._groups[iname]:
                spec = self._fetch_metric_spec(instance, metric, g)
                if spec and spec.metric_type != DynamicCheckHelper.SKIP:
                    self._metric_to_group[iname][metric] = g
                    return g

        return group

    def _fetch_metric_spec(self, instance, metric, group=None):
        """Checks whether a metric is enabled by the instance configuration

        :param instance: instance containing the check configuration
        :param metric: metric as reported from metric data source (before mapping)
        :param group: optional metric group, will be used as dot-separated prefix
        """

        instance_name = instance['name']

        # filter and classify the metric

        if group is not None:
            metric_cache = self._grp_metric_cache[instance_name].get(group, {})
            metric_map = self._grp_metric_map[instance_name].get(group, {})
            return DynamicCheckHelper._lookup_metric(metric, metric_cache, metric_map)
        else:
            metric_cache = self._metric_cache[instance_name]
            metric_map = self._metric_map[instance_name]
            return DynamicCheckHelper._lookup_metric(metric, metric_cache, metric_map)

    def is_enabled_metric(self, instance, metric, group=None):
        return self._fetch_metric_spec(
            instance, metric, group).metric_type != DynamicCheckHelper.SKIP

    def push_metric_dict(
            self,
            instance,
            metric_dict,
            labels=None,
            group=None,
            timestamp=None,
            fixed_dimensions=None,
            default_dimensions=None,
            max_depth=0,
            curr_depth=0,
            prefix='',
            index=-1):
        """This will extract metrics and dimensions from a dictionary.

        The following mappings are applied:

        Simple recursive composition of metric names:

            Input:

                {
                    'server': {
                        'requests': 12
                    }
                }

            Configuration:

                mapping:
                    counters:
                        - server_requests

            Output:

                server_requests=12

        Mapping of textual values to dimensions to distinguish array elements. Make sure that tests
        attributes are sufficient to distinguish the array elements. If not use the build-in
        'index' dimension.

            Input:

            {
                'server': [
                    {
                        'role': 'master,
                        'node_name': 'server0',
                        'requests': 1500
                    },
                    {
                        'role': 'slave',
                        'node_name': 'server1',
                        'requests': 1000
                    },
                    {
                        'role': 'slave',
                        'node_name': 'server2',
                        'requests': 500
                    }
                }
            }

            Configuration:

                mapping:
                    dimensions:
                        server_role: role
                        node_name: node_name
                    rates:
                        - requests

            Output:

                server_requests{server_role=master, node_name=server0} = 1500.0
                server_requests{server_role=slave, node_name=server1} = 1000.0
                server_requests{server_role=slave, node_name=server2} = 500.0


        Distinguish array elements where no textual attribute are available or no mapping has been
        configured for them.
        In that case an 'index' dimension will be attached to the metric which has to be mapped
        properly.

            Input:

                {
                    'server': [
                        {
                            'requests': 1500
                        },
                        {
                            'requests': 1000
                        },
                        {
                            'requests': 500
                        }
                    }
                }

            Configuration:

                mapping:
                    dimensions:
                        server_no: index          # index is a predefined label
                    counters:
                        - server_requests

            Result:

                server_requests{server_no=0} = 1500.0
                server_requests{server_no=1} = 1000.0
                server_requests{server_no=2} = 500.0


        :param instance: Instance to submit to
        :param metric_dict: input data as dictionary
        :param labels: labels to be mapped to dimensions
        :param group: group to use for mapping labels and prefixing
        :param timestamp: timestamp to report for the measurement
        :param fixed_dimensions: dimensions which are always added with fixed values
        :param default_dimensions: dimensions to be added, can be overwritten by actual data in
            metric_dict
        :param max_depth: max. depth to recurse
        :param curr_depth: depth of recursion
        :param prefix: prefix to prepend to any metric
        :param index: current index when traversing through a list
        """

        # when traversing through an array, each element must be distinguished with dimensions
        # therefore additional dimensions need to be calculated from the siblings
        # of the actual number valued fields
        if default_dimensions is None:
            default_dimensions = {}
        if fixed_dimensions is None:
            fixed_dimensions = {}
        if labels is None:
            labels = {}
        if index != -1:
            ext_labels = self.extract_dist_labels(
                instance['name'], group, metric_dict, labels.copy(), index)
            if not ext_labels:
                log.debug(
                    "skipping array due to lack of mapped dimensions for group %s "
                    "(at least 'index' should be supported)",
                    group if group else '<root>')
                return

        else:
            ext_labels = labels.copy()

        for element, child in metric_dict.items():
            # if child is a dictionary, then recurse
            if isinstance(child, dict) and curr_depth < max_depth:
                self.push_metric_dict(
                    instance,
                    child,
                    ext_labels,
                    group,
                    timestamp,
                    fixed_dimensions,
                    default_dimensions,
                    max_depth,
                    curr_depth + 1,
                    prefix + element + '_')
            # if child is a number, assume that it is a metric (it will be filtered
            # out by the rate/gauge names)
            elif isinstance(child, Number):
                self.push_metric(
                    instance,
                    prefix + element,
                    float(child),
                    ext_labels,
                    group,
                    timestamp,
                    fixed_dimensions,
                    default_dimensions)
            # if it is a list, then each array needs to be added. Additional dimensions must be
            # found in order to distinguish the measurements.
            elif isinstance(child, list):
                for i, child_element in enumerate(child):
                    if isinstance(child_element, dict):
                        if curr_depth < max_depth:
                            self.push_metric_dict(
                                instance,
                                child_element,
                                ext_labels,
                                group,
                                timestamp,
                                fixed_dimensions,
                                default_dimensions,
                                max_depth,
                                curr_depth + 1,
                                prefix + element + '_',
                                index=i)
                    elif isinstance(child_element, Number):
                        if len(self._get_mappings(instance['name'], group, 'index')) > 0:
                            idx_labels = ext_labels.copy()
                            idx_labels['index'] = str(i)
                            self.push_metric(
                                instance,
                                prefix + element,
                                float(child_element),
                                idx_labels,
                                group,
                                timestamp,
                                fixed_dimensions,
                                default_dimensions)
                        else:
                            log.debug(
                                "skipping array due to lack of mapped 'index' dimensions"
                                "for group %s",
                                group if group else '<root>')
                    else:
                        log.debug(
                            'nested arrays are not supported for configurable extraction of'
                            'element %s', element)

    def extract_dist_labels(self, instance_name, group, metric_dict, labels, index):
        """Extract additional distinguishing labels from metric dictionary. All top-level
        attributes which are strings and for which a dimension mapping is available will be
        transformed into dimensions.
        :param instance_name: instance to be used
        :param group: metric group or None for root/unspecified group
        :param metric_dict: input dictionary containing the metric at the top-level
        :param labels: labels dictionary to extend with the additional found metrics
        :param index: index value to be used as fallback if no labels can be derived from
            string-valued attributes or the derived labels are not mapped in the config.
        :return: Extended labels, already including the 'labels' passed into this method
        """
        ext_labels = None
        # collect additional dimensions first from non-metrics
        for element, child in metric_dict.items():
            if isinstance(
                child,
                str) and len(
                self._get_mappings(
                    instance_name,
                    group,
                    element)) > 0:
                if not ext_labels:
                    ext_labels = labels.copy()
                ext_labels[element] = child
        # if no additional labels supplied just take the index (if it is mapped)
        if not ext_labels and len(self._get_mappings(instance_name, group, 'index')) > 0:
            if not ext_labels:
                ext_labels = labels.copy()
            ext_labels['index'] = str(index)

        return ext_labels

    def push_metric(
            self,
            instance,
            metric,
            value,
            labels=None,
            group=None,
            timestamp=None,
            fixed_dimensions=None,
            default_dimensions=None):
        """Pushes a meter using the configured mapping information to determine metric_type
        and map the name and dimensions

        :param instance: instance containing the check configuration
        :param value: metric value (float)
        :param metric: metric as reported from metric data source (before mapping)
        :param labels: labels/tags as reported from the metric data source (before mapping)
        :param timestamp: optional timestamp to handle rates properly
        :param group: specify the metric group, otherwise it will be determined from the metric name
        :param fixed_dimensions:
        :param default_dimensions:
        """

        # determine group automatically if not specified
        if fixed_dimensions is None:
            fixed_dimensions = {}
        if labels is None:
            labels = {}
        if default_dimensions is None:
            default_dimensions = {}
        if group is None:
            group = self._get_group(instance, metric)

        metric_entry = self._fetch_metric_spec(instance, metric, group)
        if metric_entry.metric_type == DynamicCheckHelper.SKIP:
            return False

        if self._prefix:
            metric_prefix = self._prefix + '.'
        else:
            metric_prefix = ''

        if group:
            metric_prefix += group + '.'

        # determine the metric name
        metric_name = metric_prefix + metric_entry.metric_name
        # determine the target dimensions
        dims = self._map_dimensions(instance['name'], labels, group, default_dimensions)
        if dims is None:
            # regex for at least one dimension filtered the metric out
            return True

        # apply fixed default dimensions
        if fixed_dimensions:
            dims.update(fixed_dimensions)

        log.debug(
            'push %s %s = %s {%s}',
            metric_entry.metric_type,
            metric_entry.metric_name,
            value,
            dims)

        if metric_entry.metric_type == DynamicCheckHelper.RATE:
            self._check.rate(metric_name, float(value), dimensions=dims)
        elif metric_entry.metric_type == DynamicCheckHelper.GAUGE:
            self._check.gauge(metric_name, float(value), timestamp=timestamp, dimensions=dims)
        elif metric_entry.metric_type == DynamicCheckHelper.COUNTER:
            self._check.increment(metric_name, float(value), dimensions=dims)

        return True

    def get_mapped_metrics(self, instance):
        """Returns input metric names or regex for which a mapping has been defined
        :param instance: instance to consider
        :return: array of metrics
        """
        metric_list = []
        iname = instance['name']
        # collect level-0 metrics
        metric_map = self._metric_map[iname]
        metric_list.extend(metric_map.get(DynamicCheckHelper.GAUGES_KEY, []))
        metric_list.extend(metric_map.get(DynamicCheckHelper.RATES_KEY, []))
        metric_list.extend(metric_map.get(DynamicCheckHelper.COUNTERS_KEY, []))
        # collect group specific metrics
        grp_metric_map = self._grp_metric_map.get(iname, {})
        for gname, gmmap in grp_metric_map.items():
            metric_list.extend(gmmap.get(DynamicCheckHelper.GAUGES_KEY, []))
            metric_list.extend(gmmap.get(DynamicCheckHelper.RATES_KEY, []))
            metric_list.extend(gmmap.get(DynamicCheckHelper.COUNTERS_KEY, []))

        return metric_list

    def _map_dimensions(self, instance_name, labels, group, default_dimensions):
        """Transforms labels attached to input metrics into Monasca dimensions
        :param default_dimensions:
        :param group:
        :param instance_name:
        :param labels:
        :return: mapped dimensions or None if the dimensions filter did not match
            and the metric needs to be filtered
        """
        dims = default_dimensions.copy()
        #  map all specified dimension all keys
        for labelname, labelvalue in labels.items():
            mapping_arr = self._get_mappings(instance_name, group, labelname)

            target_dim = None
            for map_spec in mapping_arr:
                try:
                    # map the dimension name
                    target_dim = map_spec.dimension
                    # apply the mapping function to the value
                    if target_dim not in dims:  # do not overwrite
                        mapped_value = map_spec.map_value(labelvalue)
                        if mapped_value is None:
                            # None means: filter it out based on dimension value
                            return None
                        elif mapped_value != '':
                            dims[target_dim] = mapped_value
                            # else the dimension will not map
                except (IndexError, AttributeError):  # probably the regex was faulty
                    log.exception(
                        'dimension %s value could not be mapped from %s: regex for mapped'
                        'dimension %s does not match %s',
                        target_dim, labelvalue, labelname, map_spec.regex)
                    return None

        return dims

    def _get_mappings(self, instance_name, group, labelname):
        # obtain mappings
        # check group-specific ones first
        if group:
            mapping_arr = self._grp_dimension_map[instance_name].get(group, {}).get(labelname, [])
        else:
            mapping_arr = []
        # fall-back to global ones
        mapping_arr.extend(self._dimension_map[instance_name].get(labelname, []))
        return mapping_arr

    @staticmethod
    def _create_metric_spec(metric, metric_type, metric_map):
        """Get or create MetricSpec if metric is in list for metric_type

        :param metric: incoming metric name
        :param metric_type: GAUGE, RATE, COUNTER
        :param metric_map: dictionary with mapping configuration for a metric group or
            the entire instance
        :return: new MetricSpec entry or None if metric is not listed as metric_type
        """
        re_list = metric_map.get(DynamicCheckHelper.CONFIG_SECTIONS[metric_type], [])
        for rx in re_list:
            match_groups = re.match(rx, metric)
            if match_groups:
                metric_entry = DynamicCheckHelper.MetricSpec(
                    metric_type=metric_type,
                    metric_name=DynamicCheckHelper._normalize_metricname(
                        metric,
                        match_groups))
                return metric_entry

        return None

    @staticmethod
    def _lookup_metric(metric, metric_cache, metric_map):
        """Search cache for a MetricSpec and create if missing
        :param metric: input metric name
        :param metric_cache: cache to use
        :param metric_map: mapping config element to consider
        :return: MetricSpec for the output metric
        """
        i = DynamicCheckHelper.GAUGE
        metric_entry = metric_cache.get(metric, None)
        while not metric_entry and i < len(DynamicCheckHelper.METRIC_TYPES):
            metric_entry = DynamicCheckHelper._create_metric_spec(
                metric, DynamicCheckHelper.METRIC_TYPES[i], metric_map)
            i += 1

        if not metric_entry:
            # fall-through
            metric_entry = DynamicCheckHelper.MetricSpec(
                metric_type=DynamicCheckHelper.SKIP,
                metric_name=DynamicCheckHelper._normalize_metricname(metric))

        metric_cache[metric] = metric_entry

        return metric_entry

    @staticmethod
    def _normalize_metricname(metric, match_groups=None):
        # map metric name first
        if match_groups and match_groups.lastindex \
                and match_groups.lastindex > 0:
            metric = '_'.join(match_groups.groups())

        metric = re.sub(
            '(?!^)([A-Z]+)',
            r'_\1',
            metric.replace(
                '.',
                '_')).replace(
            '__',
            '_').lower()
        metric = re.sub(r"[,+*\-/()\[\]{}]", "_", metric)
        # Eliminate multiple _
        metric = re.sub(r"__+", "_", metric)
        # Don't start/end with _
        metric = re.sub(r"^_", "", metric)
        metric = re.sub(r"_$", "", metric)
        # Drop ._ and _.
        metric = re.sub(r"\._", ".", metric)
        metric = re.sub(r"_\.", ".", metric)

        return metric


def get_pod_dimensions(kubernetes_connector, pod_metadata, kubernetes_labels):
    pod_name = pod_metadata['name']
    pod_dimensions = {'pod_name': pod_name, 'namespace': pod_metadata['namespace']}
    if "labels" in pod_metadata:
        pod_labels = pod_metadata['labels']
        for label in kubernetes_labels:
            if label in pod_labels:
                pod_dimensions[label] = pod_labels[label]
    pod_owner_dimension_set = get_pod_owner(kubernetes_connector, pod_metadata)
    if pod_owner_dimension_set:
        pod_dimensions[pod_owner_dimension_set[0]] = pod_owner_dimension_set[1]
        pod_dimensions["owner_type"] = pod_owner_dimension_set[0]
    return pod_dimensions


def _attempt_to_get_owner_name(kubernetes_connector, resource_endpoint):
    try:
        resource = kubernetes_connector.get_request(resource_endpoint)
        resource_metadata = resource['metadata']
        pod_owner_pair = _parse_manifest_for_owner(resource_metadata)
        return None if not pod_owner_pair else pod_owner_pair[1]
    except Exception as e:
        log.info("Could not connect to api to get owner data - {}".format(e))
        return None
    return None


def _parse_manifest_for_owner(resource_metadata):
    resource_name = resource_metadata['name']
    owner_references = resource_metadata.get('ownerReferences', None)
    if owner_references:
        try:
            if len(owner_references) > 1:
                log.warn("More then one owner for resource {}".format(resource_name))
            owner_reference = owner_references[0]
            resource_owner_type = owner_reference['kind']
            resource_owner_name = owner_reference['name']
            return resource_owner_type, resource_owner_name
        except Exception:
            log.info("Could not get owner from ownerReferences"
                     "for resource {}".format(resource_name))
    # Try to get owner from annotations
    else:
        try:
            resource_created_by = json.loads(
                resource_metadata['annotations']['kubernetes.io/created-by'])
            resource_owner_type = resource_created_by['reference']['kind']
            resource_owner_name = resource_created_by['reference']['name']
            return resource_owner_type, resource_owner_name
        except Exception:
            log.info("Could not get resource owner from annotations"
                     "for resource {}".format(resource_name))
    return None


def _get_pod_owner_pair(kubernetes_connector, pod_owner_type, pod_owner_name, pod_namespace):
    if pod_owner_type == "ReplicationController":
        return 'replication_controller', pod_owner_name
    elif pod_owner_type == "ReplicaSet":
        if not kubernetes_connector:
            log.info("Can not set deployment name as connection information to API is not set. "
                     "Setting ReplicaSet as dimension")
            deployment_name = None
        else:
            replicaset_endpoint = "/apis/extensions/v1beta1/namespaces/{}/replicasets/{}".format(
                pod_namespace, pod_owner_name)
            deployment_name = _attempt_to_get_owner_name(kubernetes_connector, replicaset_endpoint)
        if not deployment_name:
            return 'replica_set', pod_owner_name
        else:
            return 'deployment', deployment_name
    elif pod_owner_type == "DaemonSet":
        return 'daemon_set', pod_owner_name
    elif pod_owner_type == "StatefulSet":
        return 'stateful_set', pod_owner_name
    elif pod_owner_type == "Job":
        if not kubernetes_connector:
            log.info("Can not set cronjob name as connection information to API is not set. "
                     "Setting job as dimension")
            cronjob_name = None
        else:
            job_endpoint = "/apis/batch/v1/namespaces/{}/jobs/{}".format(
                pod_namespace, pod_owner_name)
            cronjob_name = _attempt_to_get_owner_name(kubernetes_connector, job_endpoint)
        if not cronjob_name:
            return 'job', pod_owner_name
        else:
            return 'cronjob', cronjob_name
    else:
        log.warn("Unsupported pod owner kind {}".format(pod_owner_type))
        return None


def get_pod_owner(kubernetes_connector, pod_metadata):
    pod_namespace = pod_metadata['namespace']
    owner_pair = _parse_manifest_for_owner(pod_metadata)
    if owner_pair:
        return _get_pod_owner_pair(
            kubernetes_connector,
            owner_pair[0],
            owner_pair[1],
            pod_namespace)
    return None



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os
import re

from datetime import datetime

import monasca_agent.collector.checks as checks

log = logging.getLogger(__name__)


class Crash(checks.AgentCheck):

    def __init__(self, name, init_config, agent_config):
        super(Crash, self).__init__(name, init_config, agent_config)
        self.crash_dir = self.init_config.get('crash_dir', '/var/crash')
        log.debug('crash dir: %s', self.crash_dir)

    def _create_datetime_for_rhel71(self, dir_name):
        date_parts = dir_name.split('-')
        parts_len = len(date_parts)
        if parts_len > 2:
            try:
                # Use last two parts, because the front part(IP address part) is not checked.
                date_part = date_parts[parts_len - 2] + '-' + date_parts[parts_len - 1]
                return datetime.strptime(date_part, '%Y.%m.%d-%H:%M:%S')
            except ValueError:
                pass

    def _check_dir_name(self, dir_name):
        dt = None
        if dir_name is None:
            return None
        # Check for CentOS 7.1 and RHEL 7.1. <IP-address>-YYYY.MM.dd-HH:mm:ss
        # (e.g. 127.0.0.1-2015.10.02-16:07:51)
        elif re.match(r".*-\d{4}[.]\d{2}[.]\d{2}-\d{2}:\d{2}:\d{2}$", dir_name):
            dt = self._create_datetime_for_rhel71(dir_name)
        else:
            try:
                dt = datetime.strptime(dir_name, '%Y%m%d%H%M')
            except ValueError:
                pass
        return dt

    def check(self, instance):
        """Capture crash dump statistics
        """
        dimensions = self._set_dimensions(None, instance)
        dump_count = 0
        value_meta = None

        # Parse the crash directory
        if os.path.isdir(self.crash_dir):
            for entry in sorted(os.listdir(self.crash_dir), reverse=True):
                if os.path.isdir(os.path.join(self.crash_dir, entry)):
                    dt = self._check_dir_name(entry)
                    if dt is None:
                        continue

                    # Found a valid crash dump directory
                    log.debug('found crash dump dir: %s',
                              os.path.join(self.crash_dir, entry))
                    dump_count += 1

                    # Return the date-/timestamp of the most recent crash
                    if dump_count == 1:
                        value_meta = {'latest': str(dt)}

        log.debug('dump_count: %s', dump_count)
        self.gauge('crash.dump_count', dump_count, dimensions=dimensions,
                   value_meta=value_meta)



#!/usr/bin/env python
# coding=utf-8

# (C) Copyright 2018 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Health check will returns 0 when service is working properly."""


def main():
    """Health check for Monasca-agent collector."""
    # TODO(Dobroslaw Zybort) wait for health check endpoint ...
    return 0


if __name__ == '__main__':
    main()



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from httplib2 import Http
from httplib2 import HttpLib2Error
import json
import socket

from monasca_agent.collector.checks import AgentCheck


class Riak(AgentCheck):

    keys = ["vnode_gets",
            "vnode_puts",
            "vnode_index_reads",
            "vnode_index_writes",
            "vnode_index_deletes",
            "node_gets",
            "node_puts",
            "pbc_active",
            "pbc_connects",
            "memory_total",
            "memory_processes",
            "memory_processes_used",
            "memory_atom",
            "memory_atom_used",
            "memory_binary",
            "memory_code",
            "memory_ets",
            "read_repairs",
            "node_put_fsm_rejected_60s",
            "node_put_fsm_active_60s",
            "node_put_fsm_in_rate",
            "node_put_fsm_out_rate",
            "node_get_fsm_rejected_60s",
            "node_get_fsm_active_60s",
            "node_get_fsm_in_rate",
            "node_get_fsm_out_rate"]

    stat_keys = ["node_get_fsm_siblings",
                 "node_get_fsm_objsize",
                 "node_get_fsm_time",
                 "node_put_fsm_time"]

    def __init__(self, name, init_config, agent_config, instances=None):
        AgentCheck.__init__(self, name, init_config, agent_config, instances)
        for k in ["mean", "median", "95", "99", "100"]:
            [self.keys.append(m + "_" + k) for m in self.stat_keys]

        self.prev_coord_redirs_total = -1

    def check(self, instance):
        url = instance['url']
        default_timeout = self.init_config.get('default_timeout', 5)
        timeout = float(instance.get('timeout', default_timeout))

        dimensions = self._set_dimensions(None, instance)

        try:
            h = Http(timeout=timeout)
            resp, content = h.request(url, "GET")

        except socket.timeout:
            return

        except socket.error:
            return

        except HttpLib2Error:
            return

        stats = json.loads(content)

        [self.gauge("riak." + k, stats[k], dimensions=dimensions) for k in self.keys if k in stats]

        coord_redirs_total = stats["coord_redirs_total"]
        if self.prev_coord_redirs_total > -1:
            count = coord_redirs_total - self.prev_coord_redirs_total
            self.gauge('riak.coord_redirs', count, dimensions=dimensions)

        self.prev_coord_redirs_total = coord_redirs_total



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import defaultdict
import urllib


from monasca_agent.collector.checks import AgentCheck
from monasca_agent.common.util import headers


STATS_URL = "/;csv;norefresh"
EVENT_TYPE = SOURCE_TYPE_NAME = 'haproxy'


class Services(object):
    BACKEND = 'BACKEND'
    FRONTEND = 'FRONTEND'
    ALL = (BACKEND, FRONTEND)


class HAProxy(AgentCheck):

    def __init__(self, name, init_config, agent_config):
        AgentCheck.__init__(self, name, init_config, agent_config)

        # Host status needs to persist across all checks
        self.host_status = defaultdict(lambda: defaultdict(lambda: None))

    METRICS = {
        "qcur": ("gauge", "queue.current"),
        "scur": ("gauge", "session.current"),
        "slim": ("gauge", "session.limit"),
        "spct": ("gauge", "session.pct"),    # Calculated as: (scur/slim)*100
        "stot": ("rate", "session.rate"),
        "bin": ("rate", "bytes.in_rate"),
        "bout": ("rate", "bytes.out_rate"),
        "dreq": ("rate", "denied.req_rate"),
        "dresp": ("rate", "denied.resp_rate"),
        "ereq": ("rate", "errors.req_rate"),
        "econ": ("rate", "errors.con_rate"),
        "eresp": ("rate", "errors.resp_rate"),
        "wretr": ("rate", "warnings.retr_rate"),
        "wredis": ("rate", "warnings.redis_rate"),
        "req_rate": ("gauge", "requests.rate"),
        "hrsp_1xx": ("rate", "response.1xx"),
        "hrsp_2xx": ("rate", "response.2xx"),
        "hrsp_3xx": ("rate", "response.3xx"),
        "hrsp_4xx": ("rate", "response.4xx"),
        "hrsp_5xx": ("rate", "response.5xx"),
        "hrsp_other": ("rate", "response.other"),
    }

    def check(self, instance):
        self.dimensions = self._set_dimensions({'service': 'haproxy'}, instance)
        url = instance.get('url')
        username = instance.get('username')
        password = instance.get('password')
        collect_service_stats_only = instance.get('collect_service_stats_only', True)
        collect_aggregates_only = instance.get('collect_aggregates_only', True)
        collect_status_metrics = instance.get('collect_status_metrics', False)

        self.log.debug('Processing HAProxy data for %s' % url)

        data = self._fetch_data(url, username, password)

        self._process_data(data, collect_service_stats_only, collect_aggregates_only,
                           url=url, collect_status_metrics=collect_status_metrics)

    def _fetch_data(self, url, username, password):
        """Hit a given URL and return the parsed json.

        """
        # Try to fetch data from the stats URL

        passman = urllib.request.HTTPPasswordMgrWithDefaultRealm()
        passman.add_password(None, url, username, password)
        authhandler = urllib.request.HTTPBasicAuthHandler(passman)
        opener = urllib.request.build_opener(authhandler)
        urllib.request.install_opener(opener)
        url = "%s%s" % (url, STATS_URL)

        self.log.debug("HAProxy Fetching haproxy search data from: %s" % url)

        req = urllib.request.Request(url, None, headers(self.agent_config))
        request = urllib.request.urlopen(req)
        response = request.read()
        # Split the data by line
        return response.split('\n')

    def _process_data(self, data, collect_service_stats_only, collect_aggregates_only,
                      url=None, collect_status_metrics=False):
        """Main data-processing loop. For each piece of useful data, we'll

        save a metric.
        """

        # Split the first line into an index of fields
        # The line looks like:
        # "# pxname,svname,qcur,qmax,scur,smax,slim,stot,bin,bout,dreq,dresp,ereq,econ,eresp,wretr,
        # wredis,status,weight,act,bck,chkfail,chkdown,lastchg,downtime,qlimit,pid,iid,sid,throttle
        # ,lbtot,tracked,type,rate,rate_lim,rate_max,"
        fields = [f.strip() for f in data[0][2:].split(',') if f]

        hosts_statuses = defaultdict(int)

        # Holds a list of dictionaries describing each system
        data_list = []

        for line in data[1:]:  # Skip the first line
            if not line.strip():
                continue
            data_dict = {}
            values = line.split(',')

            # Store each line's values in a dictionary
            for i, val in enumerate(values):
                if val:
                    try:
                        # Try converting to a long, if failure, just leave it
                        val = float(val)
                    except Exception:  # nosec
                        pass
                    data_dict[fields[i]] = val

            if collect_service_stats_only and data_dict['pxname'] != 'stats':
                continue

            # The percentage of used sessions based on 'scur' and 'slim'
            if 'slim' in data_dict and 'scur' in data_dict:
                try:
                    data_dict['spct'] = (data_dict['scur'] / data_dict['slim']) * 100
                except (TypeError, ZeroDivisionError):
                    pass

            service = data_dict['svname']

            if collect_status_metrics and 'status' in data_dict and 'pxname' in data_dict:
                hosts_statuses[(data_dict['pxname'], data_dict['status'])] += 1

            if data_dict['svname'] in Services.ALL:
                data_list.append(data_dict)

                # Send the list of data to the metric callbacks
                self._process_metrics(data_list, service, url)

                # Clear out the list for the next service
                data_list = []
            elif not collect_aggregates_only:
                data_list.append(data_dict)

        if collect_status_metrics:
            self._process_status_metric(hosts_statuses)

        return data

    def _process_status_metric(self, hosts_statuses):
        agg_statuses = defaultdict(lambda: {'available': 0, 'unavailable': 0})
        status_dimensions = self.dimensions.copy()
        for (service, status), count in hosts_statuses.items():
            status = status.lower()

            status_dimensions.update({'status': status, 'component': service})
            self.gauge("haproxy.count_per_status", count, dimensions=status_dimensions)

            if 'up' in status:
                agg_statuses[service]['available'] += count
            if 'down' in status or 'maint' in status or 'nolb' in status:
                agg_statuses[service]['unavailable'] += count

        for service in agg_statuses:
            for status, count in agg_statuses[service].items():
                status_dimensions.update({'status': status, 'component': service})
                self.gauge("haproxy.count_per_status", count, dimensions=status_dimensions)

    def _process_metrics(self, data_list, service, url):
        for data in data_list:
            """Each element of data_list is a dictionary related to one host

            (one line) extracted from the csv. All of these elements should
            have the same value for 'pxname' key
            It should look like:
            data_list = [
                {'svname':'i-4562165', 'pxname':'dogweb', 'scur':'42', ...},
                {'svname':'i-2854985', 'pxname':'dogweb', 'scur':'1337', ...},
                ...
            ]
            """
            metric_dimensions = self.dimensions.copy()
            hostname = data['svname']
            service_name = data['pxname']

            metric_dimensions.update({'type': service,
                                      'instance_url': url,
                                      'component': service_name})
            if service == Services.BACKEND:
                metric_dimensions.update({'backend': hostname})

            for key, value in data.items():
                if HAProxy.METRICS.get(key):
                    suffix = HAProxy.METRICS[key][1]
                    name = "haproxy.%s.%s" % (service.lower(), suffix)
                    if HAProxy.METRICS[key][0] == 'rate':
                        self.rate(name, value, dimensions=metric_dimensions)
                    else:
                        self.gauge(name, value, dimensions=metric_dimensions)



# Copyright 2016 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import logging
from unittest import mock
import unittest
import json
import random

from monasca_agent.common import util
from monasca_agent.collector.checks_d import kibana

LOG = logging.getLogger(kibana.__name__)

_KIBANA_VERSION = '4.4.0'
_KIBANA_URL = 'http://localhost:5700/api/status'


class MockKibanaCheck(kibana.Kibana):
    def __init__(self):
        super(MockKibanaCheck, self).__init__(
            name='kibana',
            init_config={
                'url': _KIBANA_URL
            },
            instances=[],
            agent_config={}
        )


class KibanaCheckTest(unittest.TestCase):
    def setUp(self):
        super(KibanaCheckTest, self).setUp()
        with mock.patch.object(util, 'get_hostname'):
            self.kibana_check = MockKibanaCheck()
            self.kibana_check._get_kibana_version = mock.Mock(
                return_value=_KIBANA_VERSION
            )

    def test_should_throw_exception_if_url_not_specified(self):

        with self.assertRaises(Exception) as err:
            self.kibana_check.init_config = {}
            self.kibana_check.check(None)

        self.assertEqual('An url to kibana must be specified',
                         str(err.exception))

    def test_should_early_exit_if_all_metrics_disabled(self):
        with mock.patch.object(util, 'get_hostname') as _,\
                mock.patch.object(LOG, 'warning') as mock_log_warning:
            self.kibana_check._get_kibana_version = mock.Mock()
            self.kibana_check._get_data = mock.Mock()
            self.kibana_check._process_metrics = mock.Mock()

            self.kibana_check.check({'metrics': []})

            self.assertFalse(self.kibana_check._get_kibana_version.called)
            self.assertFalse(self.kibana_check._get_data.called)
            self.assertFalse(self.kibana_check._process_metrics.called)

            self.assertEqual(mock_log_warning.call_count, 1)
            self.assertEqual(mock_log_warning.call_args[0][0],
                             'All metrics have been disabled in configuration '
                             'file, nothing to do.')

    def test_failed_to_retrieve_data(self):
        with mock.patch.object(util, 'get_hostname') as _,\
                mock.patch.object(LOG, 'error') as mock_log_error,\
                mock.patch.object(LOG, 'exception') as mock_log_exception:
            exception = Exception('oh')
            self.kibana_check._get_data = mock.Mock(
                side_effect=exception)

            self.kibana_check.check({
                'metrics': ['heap_size',
                            'heap_used',
                            'load',
                            'req_sec',
                            'resp_time_avg',
                            'resp_time_max']
            })

            self.assertEqual(mock_log_error.call_count, 1)
            self.assertEqual(mock_log_error.call_args[0][0],
                             'Error while trying to get stats from Kibana[%s]'
                             % _KIBANA_URL)

            self.assertEqual(mock_log_exception.call_count, 1)
            self.assertEqual(mock_log_exception.call_args[0][0],
                             exception)

    def test_empty_data_returned(self):
        with mock.patch.object(util, 'get_hostname') as _, \
                mock.patch.object(LOG, 'warning') as mock_log_warning:
            self.kibana_check._get_data = mock.Mock(return_value=None)

            self.kibana_check.check({
                'metrics': ['heap_size',
                            'heap_used',
                            'load',
                            'req_sec',
                            'resp_time_avg',
                            'resp_time_max']
            })

            self.assertEqual(mock_log_warning.call_count, 1)
            self.assertEqual(mock_log_warning.call_args[0][0],
                             'No stats data was collected from kibana')

    def test_process_metrics(self):
        all_metrics = ['heap_size', 'heap_used', 'load',
                       'req_sec', 'resp_time_avg',
                       'resp_time_max']
        enabled_metrics = all_metrics[:random.randint(0, len(all_metrics) - 1)]

        if not enabled_metrics:
            # if random made a joke, make sure at least one metric
            # is there to check
            enabled_metrics.append(all_metrics[0])

        response = {
            'metrics': {
                'heapTotal': [],
                'heapUsed': [],
                'load': [],
                'requestsPerSecond': [],
                'responseTimeAvg': [],
                'responseTimeMax': [],
            }
        }

        with mock.patch.object(util, 'get_hostname'):
            self.kibana_check._get_data = mock.Mock(return_value=response)
            self.kibana_check._process_metric = mock.Mock()

            self.kibana_check.check({'metrics': enabled_metrics})

            self.assertTrue(self.kibana_check._process_metric.called)
            self.assertEqual(len(enabled_metrics),
                             self.kibana_check._process_metric.call_count)

    def test_check(self):
        fixture_file = os.path.dirname(
            os.path.abspath(__file__)) + '/fixtures/test_kibana.json'
        response = json.load(open(fixture_file))

        metrics = ['heap_size', 'heap_used', 'load',
                   'req_sec', 'resp_time_avg',
                   'resp_time_max']

        # expected value, see fixture values for details
        # it presents partial response kibana returns
        # mocked to always returned repeatable and known data

        # 96 values is in total
        # but 7 will be omitted because there not returned
        # in responseTimeAvg

        expected_metric = [
            'kibana.heap_size_mb',
            'kibana.heap_used_mb',
            'kibana.load_avg_1m',
            'kibana.load_avg_5m',
            'kibana.load_avg_15m',
            'kibana.req_sec',
            'kibana.resp_time_avg_ms',
            'kibana.resp_time_max_ms'
        ]

        with mock.patch.object(util, 'get_hostname'):
            self.kibana_check._get_data = mock.Mock(return_value=response)
            self.kibana_check.gauge = mock.Mock(return_value=response)

            self.kibana_check.check({'metrics': metrics})

            self.assertTrue(self.kibana_check.gauge.called)
            self.assertEqual(89, self.kibana_check.gauge.call_count)

            for call_arg in self.kibana_check.gauge.call_args_list:
                metric_name = call_arg[1]['metric']
                self.assertIn(metric_name, expected_metric)



# (C) Copyright 2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import monasca_setup.detection


class Directory(monasca_setup.detection.ServicePlugin):

    """Setup configuration to monitor directory size.
       directory_names example:
                    'directory_names': ['/path/to/directory_1',
                                        '/path/to/directory_2',
                                        ...]
       service_name example:
                    'service_name': 'directory-service'
    """

    def __init__(self, template_dir, overwrite=True, args=None):
        service_params = {
            'args': args,
            'template_dir': template_dir,
            'overwrite': overwrite,
            'service_name': '',
            'directory_names': []}
        super(Directory, self).__init__(service_params)



# Copyright 2013 Cloudbase Solutions Srl
# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
Utility class for VM related operations.
Based on the "root/virtualization/v2" namespace available starting with
Hyper-V Server / Windows Server 2012.
"""

import sys

if sys.platform == 'win32':
    import wmi

from monasca_agent.collector.virt import inspector


class HyperVException(inspector.InspectorException):
    pass


class UtilsV2(object):

    _VIRTUAL_SYSTEM_TYPE_REALIZED = 'Microsoft:Hyper-V:System:Realized'

    _PROC_SETTING = 'Msvm_ProcessorSettingData'
    _MEMORY_SETTING = "Msvm_MemorySettingData"
    _SYNTH_ETH_PORT = 'Msvm_SyntheticEthernetPortSettingData'
    _ETH_PORT_ALLOC = 'Msvm_EthernetPortAllocationSettingData'
    _PORT_ACL_SET_DATA = 'Msvm_EthernetSwitchPortAclSettingData'
    _STORAGE_ALLOC = 'Msvm_StorageAllocationSettingData'
    _VS_SETTING_DATA = 'Msvm_VirtualSystemSettingData'
    _METRICS_ME = 'Msvm_MetricForME'
    _BASE_METRICS_VALUE = 'Msvm_BaseMetricValue'

    _CPU_METRIC_NAME = 'Aggregated Average CPU Utilization'
    _MEMORY_METRIC_NAME = 'Aggregated Average Memory Utilization'
    _NET_IN_METRIC_NAME = 'Filtered Incoming Network Traffic'
    _NET_OUT_METRIC_NAME = 'Filtered Outgoing Network Traffic'
    # Disk metrics are supported from Hyper-V 2012 R2
    _DISK_RD_METRIC_NAME = 'Disk Data Read'
    _DISK_WR_METRIC_NAME = 'Disk Data Written'
    _DISK_LATENCY_METRIC_NAME = 'Average Disk Latency'
    _DISK_IOPS_METRIC_NAME = 'Average Normalized Disk Throughput'

    def __init__(self, host='.'):
        if sys.platform == 'win32':
            self._init_hyperv_wmi_conn(host)
            self._init_cimv2_wmi_conn(host)
        self._host_cpu_info = None

    def _init_hyperv_wmi_conn(self, host):
        self._conn = wmi.WMI(moniker='//%s/root/virtualization/v2' % host)

    def _init_cimv2_wmi_conn(self, host):
        self._conn_cimv2 = wmi.WMI(moniker='//%s/root/cimv2' % host)

    def get_host_cpu_info(self):
        if not self._host_cpu_info:
            host_cpus = self._conn_cimv2.Win32_Processor()
            self._host_cpu_info = (host_cpus[0].MaxClockSpeed, len(host_cpus))
        return self._host_cpu_info

    def get_all_vms(self):
        vms = [(v.ElementName, v.Name) for v in
               self._conn.Msvm_ComputerSystem(['ElementName', 'Name'],
                                              Caption="Virtual Machine")]
        return vms

    def get_cpu_metrics(self, vm_name):
        vm = self._lookup_vm(vm_name)
        cpu_sd = self._get_vm_resources(vm, self._PROC_SETTING)[0]
        cpu_metrics_def = self._get_metric_def(self._CPU_METRIC_NAME)
        cpu_metric_aggr = self._get_metrics(vm, cpu_metrics_def)

        cpu_used = 0
        if cpu_metric_aggr:
            cpu_used = int(cpu_metric_aggr[0].MetricValue)

        return (cpu_used,
                int(cpu_sd.VirtualQuantity),
                int(vm.OnTimeInMilliseconds))

    def get_memory_metrics(self, vm_name):
        vm = self._lookup_vm(vm_name)
        memory_def = self._get_metric_def(self._MEMORY_METRIC_NAME)
        metric_memory = self._get_metrics(vm, memory_def)
        memory_usage = 0
        if metric_memory:
            memory_usage = int(metric_memory[0].MetricValue)
        return memory_usage

    def get_vnic_metrics(self, vm_name):
        vm = self._lookup_vm(vm_name)
        ports = self._get_vm_resources(vm, self._ETH_PORT_ALLOC)
        vnics = self._get_vm_resources(vm, self._SYNTH_ETH_PORT)

        metric_def_in = self._get_metric_def(self._NET_IN_METRIC_NAME)
        metric_def_out = self._get_metric_def(self._NET_OUT_METRIC_NAME)

        for port in ports:
            vnic = [v for v in vnics if port.Parent == v.path_()][0]

            metric_value_instances = self._get_metric_value_instances(
                port.associators(wmi_result_class=self._PORT_ACL_SET_DATA),
                self._BASE_METRICS_VALUE)
            metric_values = self._sum_metric_values_by_defs(
                metric_value_instances, [metric_def_in, metric_def_out])

            yield {
                'rx_mb': metric_values[0],
                'tx_mb': metric_values[1],
                'element_name': vnic.ElementName,
                'address': vnic.Address
            }

    def get_disk_metrics(self, vm_name):
        vm = self._lookup_vm(vm_name)
        metric_def_r = self._get_metric_def(self._DISK_RD_METRIC_NAME)
        metric_def_w = self._get_metric_def(self._DISK_WR_METRIC_NAME)

        disks = self._get_vm_resources(vm, self._STORAGE_ALLOC)
        for disk in disks:
            metric_values = self._get_metric_values(
                disk, [metric_def_r, metric_def_w])

            # Thi sis e.g. the VHD file location
            if disk.HostResource:
                host_resource = disk.HostResource[0]

            yield {
                # Values are in megabytes
                'read_mb': metric_values[0],
                'write_mb': metric_values[1],
                'instance_id': disk.InstanceID,
                'host_resource': host_resource
            }

    def get_disk_latency_metrics(self, vm_name):
        vm = self._lookup_vm(vm_name)
        metric_latency_def = self._get_metric_def(
            self._DISK_LATENCY_METRIC_NAME)

        disks = self._get_vm_resources(vm, self._STORAGE_ALLOC)
        for disk in disks:
            metric_values = self._get_metric_values(
                disk, [metric_latency_def])

            yield {
                'disk_latency': metric_values[0],
                'instance_id': disk.InstanceID,
            }

    def get_disk_iops_count(self, vm_name):
        vm = self._lookup_vm(vm_name)
        metric_def_iops = self._get_metric_def(self._DISK_IOPS_METRIC_NAME)
        disks = self._get_vm_resources(vm, self._STORAGE_ALLOC)

        for disk in disks:
            metric_values = self._get_metric_values(
                disk, [metric_def_iops])

            yield {
                'iops_count': metric_values[0],
                'instance_id': disk.InstanceID,
            }

    @staticmethod
    def _sum_metric_values(metrics):
        tot_metric_val = 0
        for metric in metrics:
            tot_metric_val += int(metric.MetricValue)
        return tot_metric_val

    def _sum_metric_values_by_defs(self, element_metrics, metric_defs):
        metric_values = []
        for metric_def in metric_defs:
            if metric_def:
                metrics = self._filter_metrics(element_metrics, metric_def)
                metric_values.append(self._sum_metric_values(metrics))
            else:
                # In case the metric is not defined on this host
                metric_values.append(0)
        return metric_values

    @staticmethod
    def _get_metric_value_instances(elements, result_class):
        instances = []
        for el in elements:
            associators = el.associators(wmi_result_class=result_class)
            if associators:
                instances.append(associators[0])

        return instances

    def _get_metric_values(self, element, metric_defs):
        element_metrics = element.associators(
            wmi_association_class=self._METRICS_ME)
        return self._sum_metric_values_by_defs(element_metrics, metric_defs)

    def _lookup_vm(self, vm_name):
        vms = self._conn.Msvm_ComputerSystem(ElementName=vm_name)
        n = len(vms)
        if n == 0:
            raise inspector.InstanceNotFoundException(
                'VM %s not found on Hyper-V' % vm_name)
        elif n > 1:
            raise HyperVException('Duplicate VM name found: %s' % vm_name)
        else:
            return vms[0]

    def _get_metrics(self, element, metric_def):
        return self._filter_metrics(
            element.associators(
                wmi_association_class=self._METRICS_ME), metric_def)

    @staticmethod
    def _filter_metrics(all_metrics, metric_def):
        return [v for v in all_metrics if
                v.MetricDefinitionId == metric_def.Id]

    def _get_metric_def(self, metric_def):
        metric = self._conn.CIM_BaseMetricDefinition(ElementName=metric_def)
        if metric:
            return metric[0]

    def _get_vm_setting_data(self, vm):
        vm_settings = vm.associators(
            wmi_result_class=self._VS_SETTING_DATA)
        # Avoid snapshots
        return [s for s in vm_settings if
                s.VirtualSystemType == self._VIRTUAL_SYSTEM_TYPE_REALIZED][0]

    def _get_vm_resources(self, vm, resource_class):
        setting_data = self._get_vm_setting_data(vm)
        return setting_data.associators(wmi_result_class=resource_class)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import re
import subprocess

import monasca_agent.collector.checks as checks
from monasca_agent.common.psutil_wrapper import psutil


log = logging.getLogger(__name__)


class Cpu(checks.AgentCheck):

    def __init__(self, name, init_config, agent_config):
        super(Cpu, self).__init__(name, init_config, agent_config)
        process_fs_path_config = init_config.get('process_fs_path', None)
        if process_fs_path_config:
            psutil.PROCFS_PATH = process_fs_path_config
            self.log.debug('The path of the process filesystem set to %s', process_fs_path_config)
        else:
            self.log.debug('The process_fs_path not set. Use default path: /proc')
        # psutil.cpu_percent and psutil.cpu_times_percent are called in
        # __init__ because the first time these two functions are called with
        # interval = 0.0 or None, it will return a meaningless 0.0 value
        # which you are supposed to ignore.
        psutil.cpu_percent(interval=None, percpu=False)
        psutil.cpu_times_percent(interval=None, percpu=False)

    def check(self, instance):
        """Capture cpu stats
        """
        num_of_metrics = 0
        dimensions = self._set_dimensions(None, instance)

        if instance is not None:
            send_rollup_stats = instance.get("send_rollup_stats", False)
        else:
            send_rollup_stats = False

        cpu_stats = psutil.cpu_times_percent(interval=None, percpu=False)
        cpu_times = psutil.cpu_times(percpu=False)
        cpu_perc = psutil.cpu_percent(interval=None, percpu=False)

        data = {'cpu.user_perc': cpu_stats.user + cpu_stats.nice,
                'cpu.system_perc': cpu_stats.system + cpu_stats.irq + cpu_stats.softirq,
                'cpu.wait_perc': cpu_stats.iowait,
                'cpu.idle_perc': cpu_stats.idle,
                'cpu.stolen_perc': cpu_stats.steal,
                'cpu.percent': cpu_perc,
                'cpu.idle_time': cpu_times.idle,
                'cpu.wait_time': cpu_times.iowait,
                'cpu.user_time': cpu_times.user + cpu_times.nice,
                'cpu.system_time': cpu_times.system + cpu_times.irq + cpu_times.softirq}

        # Call lscpu command to get cpu frequency
        self._add_cpu_freq(data)

        for key, value in data.items():
            if data[key] is None or instance.get('cpu_idle_only') and 'idle_perc' not in key:
                continue
            self.gauge(key, value, dimensions)
            num_of_metrics += 1

        if send_rollup_stats:
            self.gauge('cpu.total_logical_cores', psutil.cpu_count(logical=True), dimensions)
            num_of_metrics += 1
        log.debug('Collected {0} cpu metrics'.format(num_of_metrics))

    def _add_cpu_freq(self, data):
        try:
            lscpu_command = subprocess.Popen(
                'lscpu', stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            lscpu_output = lscpu_command.communicate()[0].decode(
                encoding='UTF-8')
            cpu_freq_output = re.search(r"(CPU MHz:.*?(\d+\.\d+)\n)", lscpu_output)
            cpu_freq = float(cpu_freq_output.group(2))
            data['cpu.frequency_mhz'] = cpu_freq
        except Exception:
            log.exception('Cannot extract CPU MHz information using lscpu')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import sys
import inspect
import os
import signal

from monasca_agent.collector.checks import AgentCheck
from monasca_agent.common.util import Paths


def kill_subprocess(process_obj):
    try:
        process_obj.terminate()
    except AttributeError:
        os.kill(process_obj.pid, signal.SIGKILL)


def get_check(name, config_str):
    checksd_path = Paths().get_checksd_path()
    if checksd_path not in sys.path:
        sys.path.append(checksd_path)
    check_module = __import__(name)
    check_class = None
    classes = inspect.getmembers(check_module, inspect.isclass)
    for name, clsmember in classes:
        if AgentCheck in clsmember.__bases__:
            check_class = clsmember
            break
    if check_class is None:
        raise Exception(
            "Unable to import check %s. Missing a class that inherits AgentCheck" % name)

    return check_class.from_yaml(yaml_text=config_str, check_name=name)



# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
# Copyright 2017 FUJITSU LIMITED
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


import logging
import os

import monasca_setup.agent_config
import monasca_setup.detection

log = logging.getLogger(__name__)

_ZOOKEEPER_DEFAULT_CONFIG_PATH = '/etc/zookeeper/conf/zoo.cfg'
_ZOOKEEPER_DEFAULT_IP = 'localhost'
_ZOOKEEPER_DEFAULT_PORT = 2181


class Zookeeper(monasca_setup.detection.Plugin):
    """Detect Zookeeper daemons and setup configuration to monitor them.

    """

    def _detect(self):
        """Run detection, set self.available True if the service is detected.

        """
        process_found = monasca_setup.detection.find_process_cmdline('org.apache.zookeeper')
        self._cfg_file = self._get_config_file(process_found) if process_found else None
        has_config_file = self._cfg_file and os.path.isfile(self._cfg_file)

        self.available = process_found and has_config_file

        if not self.available:
            err_str = 'Plugin for Zookeeper will not be configured.'
            if not process_found:
                log.info('Zookeeper process has not been found: {0}'.format(err_str))
            elif not has_config_file:
                log.warning('Zookeeper plugin cannot find configuration '
                            'file: {0}. {1}'.format(self._cfg_file, err_str))

    def build_config(self):
        """Build the config as a Plugins object and return.

        """
        config = monasca_setup.agent_config.Plugins()
        host, port = self._read_config_file(self._cfg_file)
        # First watch the process
        log.info("\tWatching the zookeeper process.")
        config.merge(
            monasca_setup.detection.watch_process(
                ['org.apache.zookeeper.server'],
                'zookeeper',
                exact_match=False))

        log.info("\tEnabling the zookeeper plugin")
        config['zk'] = {
            'init_config': None, 'instances':
                [{'name': host, 'host': host, 'port': port, 'timeout': 3}]
        }

        return config

    def dependencies_installed(self):
        # The current plugin just does a simple socket connection to zookeeper and
        # parses the stat command
        return True  # pragma: no cover

    @staticmethod
    def _get_config_file(process):
        # Config file should be on the last place in cmdline
        cfg = process.as_dict(['cmdline'])['cmdline'][-1]
        # check if the last value in cmdline is a file
        # if not return default config file
        if os.path.isfile(cfg):
            log.debug('Found zookeeper config file: {0}'.format(cfg))
            return cfg
        log.debug('Missing zookeeper config file. Using default file: {0}'.
                  format(_ZOOKEEPER_DEFAULT_CONFIG_PATH))
        return _ZOOKEEPER_DEFAULT_CONFIG_PATH

    @staticmethod
    def _read_config_file(cfg_file):
        ip_address = _ZOOKEEPER_DEFAULT_IP
        port = _ZOOKEEPER_DEFAULT_PORT
        try:
            cfg = open(cfg_file, 'r')
            for line in cfg:
                if 'clientPortAddress' in line:
                    log.debug('Found clientPort in config file: {0}'.format(line))
                    ip_address = line.split('=')[1].strip()
                if 'clientPort' in line and 'clientPortAddress' not in line:
                    log.debug('Found clientPortAddress in config file: {0}'.format(line))
                    port = int(line.split('=')[1].strip())
        except Exception as ex:
            log.error('Failed to parse %s', cfg_file)
            log.exception(ex)
            return None
        return ip_address, port



# (C) Copyright 2017 Hewlett Packard Enterprise Development LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock
import unittest

from monasca_agent.collector.checks_d.kubernetes_api import KubernetesAPI

SUCCESS = 0
FAILURE = 1
KUBERNETES_LABELS = ['app']


class TestKubernetesAPI(unittest.TestCase):
    def setUp(self):
        unittest.TestCase.setUp(self)
        init_config = {}
        agent_config = {}
        self._kubernetes_api = KubernetesAPI('TestKubernetesAPI',
                                             init_config,
                                             agent_config)
        self._gauge = mock.Mock()
        self._kubernetes_api.gauge = self._gauge
        self._hostname = 'SUPPRESS'
        self._instance = {'derive_api_url': True}
        self._base_dimensions = {}

    def _get_api_health_check(self, instance, result_input):
        mock_check = mock.Mock(return_value=result_input)
        self._kubernetes_api._get_api_health = mock_check
        self._kubernetes_api.check(instance)

    def test_kubernetes_api_is_healthy(self):
        api_health_result = True
        self._get_api_health_check(self._instance,
                                   api_health_result)
        self._gauge.assert_called_with('kubernetes.api.health_status',
                                       SUCCESS,
                                       self._base_dimensions,
                                       hostname=self._hostname)

    def test_kubernetes_api_is_not_healthy(self):
        api_health_result = False
        self._get_api_health_check(self._instance,
                                   api_health_result)

        self._gauge.assert_called_with('kubernetes.api.health_status',
                                       FAILURE,
                                       self._base_dimensions,
                                       hostname=self._hostname)

    def _send_request(self, result_input):
        mock_check = mock.Mock(return_value=result_input)
        self._kubernetes_api._send_request = mock_check

    def test_report_cluster_component_statuses(self):
        component_statuses_request_result = {
            u'items': [
                {u'conditions': [{
                    u'status': u'True',
                    u'message': u'{"health": "true"}',
                    u'type': u'Healthy'}],
                 u'metadata': {u'creationTimestamp': None,
                               u'name': u'etcd-0'}}],
            u'kind': u'ComponentStatusList',
            u'apiVersion': u'v1',
            u'metadata': {u'selfLink': u'/api/v1/componentstatuses'}}
        self._send_request(component_statuses_request_result)
        self._kubernetes_api._report_cluster_component_statuses(
            self._base_dimensions)
        self._gauge.assert_called_with('kubernetes.component_status',
                                       SUCCESS,
                                       {'component_name': u'etcd-0'},
                                       hostname=self._hostname)

    def test_nodes_capacity_metric(self):
        nodes_request_result = {
            u'items': [
                {u'status': {
                    u'capacity': {u'cpu': u'4'},
                    u'allocatable': {},
                    u'daemonEndpoints': {
                        u'kubeletEndpoint': {u'Port': 10250}},
                    u'images': [{u'sizeBytes': 821774423,
                                 u'names': [u'image_name',
                                            u'image_name:latest']}],
                    u'conditions': [{u'status': u'False',
                                     u'type': u'OutOfDisk'}]},
                    u'metadata': {u'name': u'node01',
                                  u'uid': u'e3600619-2557-11e7-9d76-aab101'}}]}
        self._send_request(nodes_request_result)
        self._kubernetes_api._report_nodes_metrics(self._base_dimensions)
        self._gauge.assert_called_with('kubernetes.node.capacity.cpu',
                                       4.0,
                                       {'hostname': u'node01',
                                        'unit': 'cores'})

    def test_nodes_allocatable_metric(self):
        nodes_request_result = {
            u'items': [
                {u'status': {
                    u'capacity': {},
                    u'allocatable': {
                        u'alpha.kubernetes.io/nvidia-gpu': u'0',
                        u'pods': u'110'},
                    u'daemonEndpoints': {
                        u'kubeletEndpoint': {u'Port': 10250}},
                    u'images': [{u'sizeBytes': 821774423,
                                 u'names': [u'image_name',
                                            u'image_name:latest']}],
                    u'conditions': [{u'status': u'False',
                                     u'type': u'OutOfDisk'},
                                    {u'status': u'False',
                                     u'type': u'MemoryPressure'},
                                    {u'status': u'False',
                                     u'type': u'DiskPressure'},
                                    {u'status': u'True',
                                     u'type': u'Ready'}]},
                    u'metadata': {u'name': u'node01',
                                  u'uid': u'e3600619-2557-11e7-9d76-aa3201'}}]}
        self._send_request(nodes_request_result)
        self._kubernetes_api._report_nodes_metrics(self._base_dimensions)
        self._gauge.assert_called_with('kubernetes.node.allocatable.pods',
                                       110.0,
                                       {'hostname': u'node01'})

    def test_deployment_metrics(self):
        deployments_request_result = {
            u'items': [
                {u'status': {
                    u'observedGeneration': 1,
                    u'updatedReplicas': 2,
                    u'availableReplicas': 3,
                    u'replicas': 4
                    },
                 u'metadata': {u'name': u'kube-controller-manager',
                               u'labels': {
                                   u'k8s-app': u'kube-controller-manager'},
                               u'namespace': u'kube-system',
                               u'uid': u'e61835b9-2557-11e7-9d76-aabbcc201'}}]}
        self._send_request(deployments_request_result)
        self._kubernetes_api._report_deployment_metrics(self._base_dimensions,
                                                        KUBERNETES_LABELS)
        self._gauge.assert_called_with(
            'kubernetes.deployment.updated_replicas', 2,
            {'type': u'kube-controller-manager',
             'namespace': u'kube-system'},
            hostname=self._hostname)



# (C) Copyright 2015 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Detection Plugin for MK_Livestatus Nagios/Icinga Event Broker Module
   This monasca_setup detection plugin will determine if either Nagios or
   Icinga is configured on this system, if that configuration includes the
   MK_Livestatus module, if that module's configured socket path exists
   (which should be the case as long as Nagios/Icinga is running), and if
   the Monasca Agent's user can read it.  If all these criteria are met, the
   mk_livestatus Agent plugin will be enabled with a basic configuration that
   polls all services and all hosts.
"""
import logging
import monasca_setup.agent_config
import monasca_setup.detection
import os

log = logging.getLogger(__name__)

# List several common locations of Nagios/Icinga configuration files to scan
nagios_cfg_files = ['/etc/nagios/nagios.cfg',
                    '/etc/nagios/nagios.cfg',
                    '/etc/nagios3/nagios.cfg',
                    '/etc/nagios4/nagios.cfg',
                    '/usr/local/nagios/nagios.cfg',
                    '/usr/local/nagios/etc/nagios.cfg',
                    '/etc/icinga/icinga.cfg',
                    '/usr/local/icinga/icinga.cfg',
                    '/usr/local/icinga/etc/icinga.cfg']
agent_user = 'monasca-agent'


class MKLivestatus(monasca_setup.detection.Plugin):
    """Detect MK_Livestatus Nagios Event Broker and verify permissions
    """

    def _find_socket_path(self):
        """Search common Nagios/Icinga config file locations for mk_livestatus
           broker module socket path
        """
        # Search likely Nagios/Icinga config file locations
        for cfg_file in nagios_cfg_files:
            if os.path.isfile(cfg_file):
                # Search the file for the mk_livestatus socket
                with open(cfg_file) as cfg:
                    for line in cfg:
                        if line.startswith('broker_module') and line.find('mk') > 0:
                            return line.split(' ')[-1].rstrip()

    def _detect(self):
        """Set self.available=True if module is installed and path is readable
        """
        socket_path = self._find_socket_path()
        if self.dependencies_installed and socket_path is not None:
            if os.path.exists(socket_path):
                # Is it readable by the monasca-agent user?
                test_readable = os.system(
                    'sudo -u {0} ls -1 {1} >/dev/null 2>&1'.format(agent_user, socket_path))
                if test_readable != 0:
                    log.info("Not configuring MK_Livestatus:")
                    log.info("\t{0} exists but is not readable by {1}.".format(socket_path,
                                                                               agent_user))
                else:
                    self.available = True
                    self.socket_path = socket_path

    def build_config(self):
        """Build the config as a Plugins object and return.
        """
        config = monasca_setup.agent_config.Plugins()
        log.info("\tEnabling the mk_livestatus plugin")

        config['mk_livestatus'] = {'init_config': {'socket_path': self.socket_path},
                                   'instances': [{'check_type': 'service'},
                                                 {'name': 'nagios.host_status',
                                                  'check_type': 'host'}]}

        log.info("\tConfigured for all services and hosts")
        return config

    def dependencies_installed(self):
        try:
            import socket  # noqa
        except ImportError:
            return False
        else:
            return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import subprocess
import tempfile
import unittest


class TestTail(unittest.TestCase):

    def setUp(self):
        self.log_file = tempfile.NamedTemporaryFile()
        self.logrotate_config = tempfile.NamedTemporaryFile()
        self.logrotate_config.write("""%s {
            copytruncate
            notifempty
            missingok
            rotate 1
            weekly
        }""" % self.log_file.name)
        self.logrotate_config.flush()
        self.logrotate_state_file = tempfile.NamedTemporaryFile()
        self.last_line = None

    def _trigger_logrotate(self):
        subprocess.check_call([
            'logrotate',
            '-v',  # Verbose logging
            '-f',  # Force the rotation even though the file isn't old
            # Create a state file that you have file permissions for
            '-s', self.logrotate_state_file.name,
            self.logrotate_config.name
        ])

    def test_logrotate_copytruncate(self):
        from collector.checks.utils import TailFile

        def line_parser(l):
            self.last_line = l

        tail = TailFile(logging.getLogger(), self.log_file.name, line_parser)
        self.assertEqual(tail._size, 0)

        # Write some data to the log file
        init_string = "hey there, I am a log\n"
        self.log_file.write(init_string)
        self.log_file.flush()

        # Consume from the tail
        gen = tail.tail(line_by_line=False, move_end=True)
        next(gen)

        # Verify that the tail consumed the data I wrote
        self.assertEqual(tail._size, len(init_string))

        try:
            # Trigger a copytruncate logrotation on the log file
            self._trigger_logrotate()

            # Write a new line to the log file
            new_string = "I am shorter\n"
            self.log_file.write(new_string)
            self.log_file.flush()

            # Verify that the tail recognized the logrotation
            next(gen)
            self.assertEqual(self.last_line, new_string[:-1], self.last_line)
        except OSError:
            "logrotate is not present"

if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    logging.getLogger().addHandler(logging.StreamHandler())
    unittest.main()



# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development Company LP
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import logging
import threading

import monasca_agent.common.emitter as emitter
import monasca_agent.common.util as util

log = logging.getLogger(__name__)


# Since we call flush more often than the metrics aggregation interval, we should
#  log a bunch of flushes in a row every so often.
FLUSH_LOGGING_PERIOD = 70
FLUSH_LOGGING_INITIAL = 10
FLUSH_LOGGING_COUNT = 5
EVENT_CHUNK_SIZE = 50


class Reporter(threading.Thread):
    """The reporter periodically sends the aggregated metrics to the
    server.
    """

    def __init__(self, interval, aggregator, api_host, event_chunk_size=None):
        threading.Thread.__init__(self)
        self.interval = int(interval)
        self.finished = threading.Event()
        self.aggregator = aggregator
        self.flush_count = 0
        self.log_count = 0

        self.api_host = api_host
        self.event_chunk_size = event_chunk_size or EVENT_CHUNK_SIZE

    @staticmethod
    def serialize_metrics(metrics):
        return json.dumps({"series": metrics})

    def stop(self):
        log.info("Stopping reporter")
        self.finished.set()

    def run(self):

        log.info("Reporting to %s every %ss" % (self.api_host, self.interval))

        while not self.finished.isSet():  # Use camel case isSet for 2.4 support.
            self.finished.wait(self.interval)
            self.flush()

        # Clean up the status messages.
        log.debug("Stopped reporter")

    def flush(self):
        try:
            self.flush_count += 1
            self.log_count += 1

            metrics = self.aggregator.flush()
            count = len(metrics)
            if self.flush_count % FLUSH_LOGGING_PERIOD == 0:
                self.log_count = 0
            if count:
                try:
                    emitter.http_emitter(metrics, log, self.api_host)
                except Exception:
                    log.exception("Error running emitter.")

            should_log = self.flush_count <= FLUSH_LOGGING_INITIAL \
                or self.log_count <= FLUSH_LOGGING_COUNT
            log_func = log.info
            if not should_log:
                log_func = log.debug
            log_func(
                "Flush #%s: flushed %s metric%s" %
                (self.flush_count,
                 count,
                 util.plural(count)))
            if self.flush_count == FLUSH_LOGGING_INITIAL:
                log.info(
                    "First flushes done, %s flushes will be logged every %s flushes." %
                    (FLUSH_LOGGING_COUNT, FLUSH_LOGGING_PERIOD))
        except Exception:
            log.exception("Error flushing metrics")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest
import logging

logger = logging.getLogger(__file__)

from tests.common import get_check
from nose.plugins.skip import SkipTest


class TestWeb(unittest.TestCase):

    def setUp(self):
        self.apache_config = """
init_config:

instances:
    -   apache_status_url: http://localhost:9444/server-status
        dimensions:
            instance:first
    -   apache_status_url: http://localhost:9444/server-status?auto
        dimensions:
            instance:second
"""

        self.nginx_config = """
init_config:

instances:
    -   nginx_status_url: http://localhost:44441/nginx_status/
    -   nginx_status_url: http://localhost:44441/nginx_status/
        dimensions:
            test:first_one
    -   nginx_status_url: http://dummyurl:44441/nginx_status/
        dimensions:
            test1:dummy
    -   nginx_status_url: http://localhost:44441/nginx_status/
        dimensions:
            test2:second
"""

        self.lighttpd_config = """
init_config:

instances:
    -   lighttpd_status_url: http://localhost:9449/server-status
        dimensions:
            instance:first
    -   lighttpd_status_url: http://localhost:9449/server-status?auto
        dimensions:
            instance:second
"""

    def testApache(self):
        raise SkipTest("Requires running apache")
        a, instances = get_check('apache', self.apache_config)

        a.check(instances[0])
        metrics = a.get_metrics()
        self.assertEqual(metrics[0][3].get('dimensions'), {'instance': 'first'})

        a.check(instances[1])
        metrics = a.get_metrics()
        self.assertEqual(metrics[0][3].get('dimensions'), {'instance': 'second'})

    def testApacheOldConfig(self):
        a, _ = get_check('apache', self.apache_config)
        config = {
            'apache_status_url': 'http://example.com/server-status?auto'
        }
        instances = a.parse_agent_config(config)['instances']
        assert instances[0]['apache_status_url'] == config['apache_status_url']

    def testNginx(self):
        raise SkipTest("Requires running Lighthttpd")
        nginx, instances = get_check('nginx', self.nginx_config)
        nginx.check(instances[0])
        r = nginx.get_metrics()
        self.assertEqual(len([t for t in r if t[0] == "nginx.net.connections"]), 1, r)

        nginx.check(instances[1])
        r = nginx.get_metrics()
        self.assertEqual(r[0][3].get('dimensions'), {'test': 'first_one'})

    def testNginxOldConfig(self):
        nginx, _ = get_check('nginx', self.nginx_config)
        config = {
            'nginx_status_url_1': 'http://www.example.com/nginx_status:first_tag',
            'nginx_status_url_2': 'http://www.example2.com/nginx_status:8080:second_tag',
            'nginx_status_url_3': 'http://www.example3.com/nginx_status:third_tag'
        }
        instances = nginx.parse_agent_config(config)['instances']
        self.assertEqual(len(instances), 3)
        for i, instance in enumerate(instances):
            assert ':'.join(config.values()[i].split(':')[:-1]) == instance['nginx_status_url']

    def testLighttpd(self):
        raise SkipTest("Requires running Lighthttpd")
        l, instances = get_check('lighttpd', self.lighttpd_config)

        l.check(instances[0])
        metrics = l.get_metrics()
        self.assertEqual(metrics[0][3].get('dimensions'), {'instance': 'first'})

        l.check(instances[1])
        metrics = l.get_metrics()
        self.assertEqual(metrics[0][3].get('dimensions'), {'instance': 'second'})


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# (C) Copyright 2015,2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""
    Licensed under Simplified BSD License (see LICENSE)
    (C) Boxed Ice 2010 all rights reserved
    (C) Datadog, Inc. 2010-2013 all rights reserved
"""

# Standard imports
import logging
import signal
import socket
import sys

# set up logging before importing any other components
import monasca_agent.common.util as util
util.initialize_logging('forwarder')

import os
os.umask(0o22)

# Tornado
import tornado.escape
import tornado.httpclient
import tornado.httpserver
import tornado.ioloop
import tornado.options
import tornado.web

# agent import
import monasca_agent.common.config as cfg
import monasca_agent.forwarder.api.monasca_api as mon


log = logging.getLogger('forwarder')

# Max amount of iterations to wait to meet min batch size before flushing
MAX_FLUSH_ATTEMPTS = 3

MIN_BATCH_SIZE = 200

message_batch = []

# In seconds
FLUSH_INTERVAL = 1


class AgentInputHandler(tornado.web.RequestHandler):
    def post(self):
        """Read the message and add it to the batch.
            Batch will be sent to Monasca API once the batch size or max wait time
            has been reached. Whichever one first.
        """
        global message_batch

        try:
            msg = tornado.escape.json_decode(self.request.body)
            message_batch.extend(msg)
        except Exception:
            log.exception('Error parsing body of Agent Input')
            raise tornado.web.HTTPError(500)


class Forwarder(tornado.web.Application):
    def __init__(self, port, agent_config, skip_ssl_validation=False,
                 use_simple_http_client=False):

        self._unflushed_iterations = 0
        self._endpoint = mon.MonascaAPI(agent_config)

        self._ioloop = tornado.ioloop.IOLoop.instance()

        self._port = int(port)
        self._flush_interval = FLUSH_INTERVAL * 1000
        self._non_local_traffic = agent_config.get("non_local_traffic", False)

        logging.getLogger().setLevel(agent_config.get('log_level', logging.INFO))

        self.skip_ssl_validation = skip_ssl_validation or agent_config.get(
            'skip_ssl_validation', False)
        self.use_simple_http_client = use_simple_http_client
        if self.skip_ssl_validation:
            log.info("Skipping SSL hostname validation, useful when using a transparent proxy")

    def log_request(self, handler):
        """Override the tornado logging method.
        If everything goes well, log level is DEBUG.
        Otherwise it's WARNING or ERROR depending on the response code.
        """
        if handler.get_status() < 400:
            log_method = log.debug
        elif handler.get_status() < 500:
            log_method = log.warning
        else:
            log_method = log.error
        request_time = 1000.0 * handler.request.request_time()
        log_method("%d %s %.2fms", handler.get_status(),
                   handler._request_summary(), request_time)

    def _add_tornado_handlers(self):
        handlers = [
            (r"/intake/?", AgentInputHandler)
        ]

        settings = dict(  # nosec B106
            cookie_secret="12oETzKXQAGaYdkL5gEmGeJJFuYh7EQnp2XdTP1o/Vo=",
            xsrf_cookies=False,
            debug=False,
            log_function=self.log_request
        )

        tornado.web.Application.__init__(self, handlers, **settings)

    def _bind_http_server(self, http_server):
        try:
            # non_local_traffic must be == True to match, not just some non-false value
            if self._non_local_traffic is True:
                http_server.listen(self._port)
            else:
                # localhost in lieu of 127.0.0.1 to support IPv6
                try:
                    http_server.listen(self._port, address="localhost")
                except socket.gaierror:
                    log.warning(
                        "localhost seems undefined in your host file, using 127.0.0.1 instead")
                    http_server.listen(self._port, address="127.0.0.1")
                except socket.error as e:
                    if "Errno 99" in str(e):
                        log.warning("IPv6 doesn't seem to be fully supported. Falling back to IPv4")
                        http_server.listen(self._port, address="127.0.0.1")
                    else:
                        raise
        except socket.error as e:
            log.exception(
                "Socket error %s. Is another application listening on the same port ? Exiting", e)
            sys.exit(1)
        except Exception:
            log.exception("Uncaught exception. Forwarder is exiting.")
            sys.exit(1)

        log.info("Listening on port %d" % self._port)

    def _post_metrics(self):
        global message_batch
        self._endpoint.post_metrics(message_batch)
        log.info("wrote {}".format(len(message_batch)))
        message_batch = []
        self._unflushed_iterations = 0

    def flush(self):
        if not message_batch:
            return
        if len(message_batch) >= MIN_BATCH_SIZE or self._unflushed_iterations >= MAX_FLUSH_ATTEMPTS:
            self._post_metrics()
        else:
            self._unflushed_iterations += 1

    def run(self):
        log.info("Forwarder RUN")
        self._add_tornado_handlers()

        http_server = tornado.httpserver.HTTPServer(self)
        self._bind_http_server(http_server)

        callback = tornado.ioloop.PeriodicCallback(self.flush,
                                                   self._flush_interval)

        callback.start()

        self._ioloop.start()

    def stop(self):
        self._ioloop.stop()
        log.info("Stopped")


def init_forwarder(skip_ssl_validation=False, use_simple_http_client=False):
    config = cfg.Config()
    agent_config = config.get_config(['Main', 'Api', 'Logging'])

    port = agent_config['listen_port']
    if port is None:
        port = 17123
    else:
        port = int(port)

    app = Forwarder(port, agent_config, skip_ssl_validation=skip_ssl_validation,
                    use_simple_http_client=use_simple_http_client)

    def sigterm_handler(signum, frame):
        log.info("caught sigterm. stopping")
        app.stop()

    signal.signal(signal.SIGTERM, sigterm_handler)
    signal.signal(signal.SIGINT, sigterm_handler)

    return app


def main():
    tornado.options.define("sslcheck", default=1, help="Verify SSL hostname, on by default")
    tornado.options.define("use_simple_http_client", default=0,
                           help="Use Tornado SimpleHTTPClient instead of CurlAsyncHTTPClient")
    tornado.options.define("config_file", default=None,
                           help="Location for an alternate config rather than "
                                "using the default config location.")
    args = tornado.options.parse_command_line()
    skip_ssl_validation = False
    use_simple_http_client = False

    if str(tornado.options.options.sslcheck) == u"0":
        skip_ssl_validation = True

    if str(tornado.options.options.use_simple_http_client) == u"1":
        use_simple_http_client = True

    # If we don't have any arguments, run the server.
    if not args:
        app = init_forwarder(skip_ssl_validation, use_simple_http_client=use_simple_http_client)
        app.run()

    else:
        usage = "%s [help|info]. Run with no commands to start the server" % (sys.argv[0])
        command = args[0]
        if command == 'help':
            print(usage)
        else:
            print("Unknown command: %s" % command)
            print(usage)
            return -1
    return 0


if __name__ == "__main__":
    sys.exit(main())



#
# Copyright 2012 Red Hat, Inc
# (C) Copyright 2015-2016 Hewlett Packard Enterprise Development LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Implementation of Inspector abstraction for libvirt."""

import logging

from lxml import etree
import os
from oslo_utils import units

from monasca_agent.collector.virt import inspector as virt_inspector
import monasca_agent.common.config as configuration
import monasca_agent.common.util as util

libvirt = None

log = logging.getLogger(__name__)


# Relevant configuration options must be set in 'conf.d/libvirt.yaml'
# file under 'init_config'.
# eg.
# init_config:
#   libvirt_type: kvm
#   libvirt_uri: qemu:///system
#
# 'libvirt_type' valid choices: ('kvm', 'lxc', 'qemu', 'uml', 'xen')
# 'libvirt_uri' is dependent on 'libvirt_type'

paths = util.Paths()
conf_path = os.path.join(paths.get_confd_path(), 'libvirt.yaml')

config = configuration.Config()

if os.path.exists(conf_path):
    try:
        check_config = config.check_yaml(conf_path)
    except Exception as e:
        log.exception('Unable to parse yaml config in {}\n{}'
                      .format(conf_path, str(e)))
else:
    log.debug('No conf.d/libvirt.yaml found for virt/libvirt/inspector.py')

init_config = check_config.get('init_config', {})
if init_config is None:
    init_config = {}


def retry_on_disconnect(function):
    def decorator(self, *args, **kwargs):
        try:
            return function(self, *args, **kwargs)
        except libvirt.libvirtError as e:
            if (e.get_error_code() == libvirt.VIR_ERR_SYSTEM_ERROR and
                e.get_error_domain() in (libvirt.VIR_FROM_REMOTE,
                                         libvirt.VIR_FROM_RPC)):
                self.connection = None
                return function(self, *args, **kwargs)
            else:
                raise
    return decorator


class LibvirtInspector(virt_inspector.Inspector):

    per_type_uris = dict(uml='uml:///system', xen='xen:///', lxc='lxc:///')

    def __init__(self):
        self.uri = self._get_uri()
        self.connection = None

    def _get_uri(self):
        return init_config.get('libvirt_uri') or \
            self.per_type_uris.get(init_config.get('libvirt_type'),
                                   'qemu:///system')

    def _get_connection(self):
        if not self.connection:
            global libvirt
            if libvirt is None:
                libvirt = __import__('libvirt')
            self.connection = libvirt.openReadOnly(self.uri)

        return self.connection

    @retry_on_disconnect
    def _lookup_by_uuid(self, instance):
        instance_name = instance.name()
        try:
            return self._get_connection().lookupByUUIDString(instance.UUIDString())
        except Exception as ex:
            if not libvirt or not isinstance(ex, libvirt.libvirtError):
                raise virt_inspector.InspectorException(str(ex))
            error_code = ex.get_error_code()
            if (error_code == libvirt.VIR_ERR_SYSTEM_ERROR and
                ex.get_error_domain() in (libvirt.VIR_FROM_REMOTE,
                                          libvirt.VIR_FROM_RPC)):
                raise
            msg = ("Error from libvirt while looking up instance "
                   "<name=%(name)s, id=%(id)s>: "
                   "[Error Code %(error_code)s] "
                   "%(ex)s") % {'name': instance_name,
                                'id': instance.ID(),
                                'error_code': error_code,
                                'ex': ex}
            raise virt_inspector.InstanceNotFoundException(msg)

    def inspect_cpus(self, instance):
        domain = self._lookup_by_uuid(instance)
        dom_info = domain.info()
        return virt_inspector.CPUStats(number=dom_info[3], time=dom_info[4])

    def _get_domain_not_shut_off_or_raise(self, instance):
        instance_name = instance.name()
        domain = self._lookup_by_uuid(instance)

        state = domain.info()[0]
        if state == libvirt.VIR_DOMAIN_SHUTOFF:
            msg = ('Failed to inspect data of instance '
                   '<name=%(name)s, id=%(id)s>, '
                   'domain state is SHUTOFF.') % {
                'name': instance_name, 'id': instance.ID()}
            raise virt_inspector.InstanceShutOffException(msg)

        return domain

    def inspect_vnics(self, instance):
        domain = self._get_domain_not_shut_off_or_raise(instance)

        tree = etree.fromstring(domain.XMLDesc(0))
        for iface in tree.findall('devices/interface'):
            target = iface.find('target')
            if target is not None:
                name = target.get('dev')
            else:
                continue
            mac = iface.find('mac')
            if mac is not None:
                mac_address = mac.get('address')
            else:
                continue
            fref = iface.find('filterref')
            if fref is not None:
                fref = fref.get('filter')

            params = dict((p.get('name').lower(), p.get('value'))
                          for p in iface.findall('filterref/parameter'))
            interface = virt_inspector.Interface(name=name, mac=mac_address,
                                                 fref=fref, parameters=params)
            dom_stats = domain.interfaceStats(name)
            stats = virt_inspector.InterfaceStats(rx_bytes=dom_stats[0],
                                                  rx_packets=dom_stats[1],
                                                  rx_errors=dom_stats[2],
                                                  rx_dropped=dom_stats[3],
                                                  tx_bytes=dom_stats[4],
                                                  tx_packets=dom_stats[5],
                                                  tx_errors=dom_stats[6],
                                                  tx_dropped=dom_stats[7])
            yield (interface, stats)

    def inspect_disks(self, instance):
        domain = self._get_domain_not_shut_off_or_raise(instance)

        tree = etree.fromstring(domain.XMLDesc(0))
        for device in filter(
                bool,
                [target.get("dev")
                 for target in tree.findall('devices/disk/target')]):
            disk = virt_inspector.Disk(device=device)
            block_stats = domain.blockStats(device)
            stats = virt_inspector.DiskStats(read_requests=block_stats[0],
                                             read_bytes=block_stats[1],
                                             write_requests=block_stats[2],
                                             write_bytes=block_stats[3],
                                             errors=block_stats[4])
            yield (disk, stats)

    def inspect_memory_usage(self, instance, duration=None):
        instance_name = instance.name()
        domain = self._get_domain_not_shut_off_or_raise(instance)

        try:
            memory_stats = domain.memoryStats()
            if (memory_stats and
                    memory_stats.get('available') and
                    memory_stats.get('unused')):
                memory_used = (memory_stats.get('available') -
                               memory_stats.get('unused'))
                # Stat provided from libvirt is in KB, converting it to MB.
                memory_used = memory_used / units.Ki
                return virt_inspector.MemoryUsageStats(usage=memory_used)
            else:
                msg = ('Failed to inspect memory usage of instance '
                       '<name=%(name)s, id=%(id)s>, '
                       'can not get info from libvirt.') % {
                    'name': instance_name, 'id': instance.ID()}
                raise virt_inspector.NoDataException(msg)
        # memoryStats might launch an exception if the method is not supported
        # by the underlying hypervisor being used by libvirt.
        except libvirt.libvirtError as e:
            msg = ('Failed to inspect memory usage of %(instance_uuid)s, '
                   'can not get info from libvirt: %(error)s') % {
                'instance_uuid': domain.ID(), 'error': e}
            raise virt_inspector.NoDataException(msg)

    def inspect_disk_info(self, instance):
        domain = self._get_domain_not_shut_off_or_raise(instance)

        tree = etree.fromstring(domain.XMLDesc(0))
        for disk in tree.findall('devices/disk'):
            disk_type = disk.get('type')
            if disk_type:
                if disk_type == 'network':
                    log.debug('Inspection disk usage of network disk '
                              '%(instance_uuid)s unsupported by libvirt' % {
                                  'instance_uuid': instance.ID()})
                    continue
                target = disk.find('target')
                device = target.get('dev')
                if device:
                    dsk = virt_inspector.Disk(device=device)
                    block_info = domain.blockInfo(device)
                    info = virt_inspector.DiskInfo(capacity=block_info[0],
                                                   allocation=block_info[1],
                                                   physical=block_info[2])
                    yield (dsk, info)

    def inspect_memory_resident(self, instance, duration=None):
        domain = self._get_domain_not_shut_off_or_raise(instance)
        memory = domain.memoryStats()['rss'] / units.Ki
        return virt_inspector.MemoryResidentStats(resident=memory)
